# -*- coding: utf-8 -*-
"""cheb.py

This module defines a container for Chebyshev polynomials, stored as a numba
jitclass.
"""

from collections import OrderedDict
from numba.experimental import jitclass
from numba import float64, int64
import numpy as np

from ..spectral.cheb import cheb

# pylint: disable=C0103

Cheb_spec = OrderedDict()
Cheb_spec['N'] = int64[::1]       # degrees of sets of the polynomial
Cheb_spec['x'] = float64[:, ::1]  # collocation points as arrays
# Chebyshev polynomials
Cheb_spec['Tn'] = float64[:, :, ::1]
Cheb_spec['dTn'] = float64[:, :, ::1]   # first derivative
Cheb_spec['d2Tn'] = float64[:, :, ::1]  # second derivatives


@jitclass(Cheb_spec)
class ChebContainer():
    """A container for chebyshev matrices.

    ChebContainer contains chebyshev matrices for different degrees of
    approximation.

    Attributes:
        N (int[:]): Array of the degrees of the chebyshev approximations.
        x (float[:,:]): Array such that x[:N[k], k] contains the collocation
                        points for the chebyshev approximation at iterate k.
        Tn (float[:,:,:]): Array such that Tn[:N[k], :N[k], k] contains the
                           chebyshev polynomials evaluated at the collocation
                           points for iterate k.
        dTn (float[:,:,:]): Array such that dTn[:N[k], :N[k], k] contains the
                            derivative of the chebyshev polynomials evaluated
                            at the collocation points for iterate k.
        d2Tn (float[:,:,:]): Array such that d2Tn[:N[k], :N[k], k] contains the
                             second derivative of the chebyshev polynomials
                             evaluated at the collocation points for iterate k.
    """

    def __init__(self, parameters):
        chebIts = parameters.solver.chebIts
        maxN = parameters.solver.maxN
        self.N = np.zeros((chebIts), dtype=np.int64)
        self.x = np.zeros((maxN, chebIts), dtype=np.float64)
        self.Tn = np.zeros((maxN, maxN, chebIts), dtype=np.float64)
        self.dTn = np.zeros((maxN, maxN, chebIts), dtype=np.float64)
        self.d2Tn = np.zeros((maxN, maxN, chebIts), dtype=np.float64)

        minN_log2 = parameters.solver.minN_log2
        for k in range(0, chebIts):
            N = 2**(k+minN_log2)
            self.N[k] = np.int64(N)
            x, T, dT, d2T = cheb(N)
            self.x[:N, k] = x
            self.Tn[:N, :N, k] = T
            self.dTn[:N, :N, k] = dT
            self.d2Tn[:N, :N, k] = d2T

    def get_cheb(self, k):
        """Get Chebyshev matrices.

        This function gets chebyshev matrices from the ChebContainter.

        Args:
            k (int): the index of the iterates.

        Raises:
            ValueError: if the index is negative or greater than N.size
        """
        if k < 0 or k > self.N.size:
            raise ValueError(
                'In VelocityContainer upper_z, k must be in the range '
                + '[0, parameters.solver.chebIts].')
        N = self.N[k]
        x = self.x[:N, k]
        Tn = self.Tn[:N, :N, k]
        dTn = self.dTn[:N, :N, k]
        d2Tn = self.d2Tn[:N, :N, k]
        return x, Tn, dTn, d2Tn


# pylint: disable=E1101
ChebContainer_type = ChebContainer.class_type.instance_type



from numpy import (pi, arange, hstack)

def grid_freq(num_points):
    """Constructs the coordinates and wavenumbers for a fourier grid on [-pi,pi]

    Args:
        num_points: the number of points (should be a power of 2)

    Returns:
        x: the coordinates as a 1d numpy array
        k: the wavenumbers as a 1d numpy array
    """
    x = -pi + arange(0, num_points)*2*pi/num_points
    delta_k = 1
    k = delta_k*hstack((arange(0, num_points/2), arange(-num_points/2, 0)))
    return x, k



from collections import OrderedDict

import numpy as np
from numba import float64, int64, optional
from numba.experimental import jitclass
from numba.types import unicode_type

source_spec = OrderedDict()
source_spec["name"] = optional(unicode_type)  # name of source
source_spec["latitude"] = float64  # latitude of source
source_spec["longitude"] = float64  # longitude of source
source_spec["utmcode"] = int64  # EPSG utm code of source
source_spec["radius"] = float64  # radius of the source
source_spec["PlumeHeight"] = float64  # height of the source plume
source_spec["MER"] = float64  # mass eruption rate
source_spec["duration"] = float64  # duration of eruption
source_spec["Suzuki_k"] = optional(float64)  # Suzuki k parameter
source_spec["Suzuki_peak"] = optional(float64)  # Suzuki_peak parameter


@jitclass(source_spec)
class SourceParameters:
    def __init__(
        self,
        lat,
        lon,
        utmcode,
        radius=10e3,
        PlumeHeight=10e3,
        MER=1e6,
        duration=18000,
        Suzuki_k=None,
        Suzuki_peak=0.9,
        name=None,
    ):

        self.latitude = np.float64(lat)
        self.longitude = np.float64(lon)
        self.utmcode = utmcode

        if radius < 0:
            raise ValueError("In SourceParameters, radius must be positive")
        self.radius = np.float64(radius)

        if PlumeHeight < 0:
            raise ValueError("In SourceParameters, PlumeHeight must be positive")
        self.PlumeHeight = np.float64(PlumeHeight)

        if MER < 0:
            raise ValueError("In SourceParameters, MER must be positive")
        self.MER = np.float64(MER)

        if duration < 0:
            raise ValueError("In SourceParameters, duration must be positive")
        self.duration = np.float64(duration)

        if Suzuki_k is not None:
            if Suzuki_k < 1:
                raise ValueError("In SourceParameters, must have Suzuki_k>=1")
            self.Suzuki_k = np.float64(Suzuki_k)
            self.peak_from_k()

        if Suzuki_peak is not None:
            if Suzuki_peak < 0 or Suzuki_peak > 1:
                raise ValueError("In SourceParameters, must have 0<Suzuki_peak<=1")
            self.Suzuki_peak = Suzuki_peak
            self.k_from_peak()

        if name is not None:
            self.name = name

    def k_from_peak(self):
        self.Suzuki_k = 1.0 / (1.0 - self.Suzuki_peak)

    def peak_from_k(self):
        self.Suzuki_peak = 1.0 - 1.0 / self.Suzuki_k

    def describe(self):
        print("Source parameters for AshDisperse")
        print("  Mass eruption rate MER = ", self.MER, " kg/s")
        print("  Eruption duration = ", self.duration, " s")
        print("  Plume height H = ", self.PlumeHeight, " m")
        print("  Gaussian source radius = ", self.radius, " m")
        print("  Suzuki emission profile k-parameter = ", self.Suzuki_k)
        print("  Suzuki emission profile peak-parameter = ", self.Suzuki_peak)
        print("********************")


# pylint: disable=E1101
SourceParameters_type = SourceParameters.class_type.instance_type



# from numba.pycc import CC
import numpy as np
from numba import float64, int64, jit

from ..containers.velocities import VelocityContainer_type

# cc = CC('getters')
# cc.verbose = True


# @cc.export("lower_z", float64[::1](VelocityContainer_type, int64))
@jit(
    float64[::1](VelocityContainer_type, int64),
    nopython=True,
    cache=True,
    fastmath=True,
)
def lower_z(velocities, k):
    """ """
    if k < 0 or k >= velocities.N.size:
        raise ValueError(
            "In lower_z, k must be in the range " + "[0, parameters.solver.chebIts]."
        )
    N = velocities.N[k]
    z = np.empty((N), dtype=np.float64)
    z[:N] = velocities.z[:N, 0, k]
    return z


# @cc.export('upper_z', float64[::1](VelocityContainer_type, int64))
@jit(
    float64[::1](VelocityContainer_type, int64),
    nopython=True,
    cache=True,
    fastmath=True,
)
def upper_z(velocities, k):
    if k < 0 or k >= velocities.N.size:
        raise ValueError(
            "In upper_z, k must be in the range " + "[0, parameters.solver.chebIts]."
        )
    N = velocities.N[k] - 1
    z = np.empty((N), dtype=np.float64)
    z[:N] = velocities.z[:N, 1, k]
    return z


# @cc.export('lower_U', float64[::1](VelocityContainer_type, int64))
@jit(
    float64[::1](VelocityContainer_type, int64),
    nopython=True,
    cache=True,
    fastmath=True,
)
def lower_U(velocities, k):
    if k < 0 or k >= velocities.N.size:
        raise ValueError(
            "In lower_U, k must be in the range " + "[0, parameters.solver.chebIts]."
        )
    N = velocities.N[k]
    u = np.empty((N), dtype=np.float64)
    u[:N] = velocities.U[:N, 0, k]
    return u


# @cc.export('upper_U', float64[::1](VelocityContainer_type, int64))
@jit(
    float64[::1](VelocityContainer_type, int64),
    nopython=True,
    cache=True,
    fastmath=True,
)
def upper_U(velocities, k):
    if k < 0 or k >= velocities.N.size:
        raise ValueError(
            "In upper_U, k must be in the range " + "[0, parameters.solver.chebIts]."
        )
    N = velocities.N[k] - 1
    u = np.empty((N), dtype=np.float64)
    u[:N] = velocities.U[:N, 1, k]
    return u


# @cc.export('lower_V', float64[::1](VelocityContainer_type, int64))
@jit(
    float64[::1](VelocityContainer_type, int64),
    nopython=True,
    cache=True,
    fastmath=True,
)
def lower_V(velocities, k):
    if k < 0 or k >= velocities.N.size:
        raise ValueError(
            "In lower_V, k must be in the range " + "[0, parameters.solver.chebIts]."
        )
    N = velocities.N[k]
    v = np.empty((N), dtype=np.float64)
    v[:N] = velocities.V[:N, 0, k]
    return v


# @cc.export('upper_V', float64[::1](VelocityContainer_type, int64))
@jit(
    float64[::1](VelocityContainer_type, int64),
    nopython=True,
    cache=True,
    fastmath=True,
)
def upper_V(velocities, k):
    if k < 0 or k >= velocities.N.size:
        raise ValueError(
            "In upper_V, k must be in the range " + "[0, parameters.solver.chebIts]."
        )
    N = velocities.N[k] - 1
    v = np.empty((N), dtype=np.float64)
    v[:N] = velocities.V[:N, 1, k]
    return v


# @cc.export('lower_Ws', float64[:, ::1](VelocityContainer_type, int64))
@jit(
    float64[:, ::1](VelocityContainer_type, int64),
    nopython=True,
    cache=True,
    fastmath=True,
)
def lower_Ws(velocities, k):
    if k < 0 or k >= velocities.N.size:
        raise ValueError(
            "In lower_Ws, k must be in the range " + "[0, parameters.solver.chebIts]."
        )
    N = velocities.N[k]
    ws = np.empty((N, velocities.Ws.shape[3]), dtype=np.float64)
    ws[:N, :] = velocities.Ws[:N, 0, k, :]
    return ws


# @cc.export('upper_Ws', float64[:, ::1](VelocityContainer_type, int64))
@jit(
    float64[:, ::1](VelocityContainer_type, int64),
    nopython=True,
    cache=True,
    fastmath=True,
)
def upper_Ws(velocities, k):
    if k < 0 or k >= velocities.N.size:
        raise ValueError(
            "In upper_Ws, k must be in the range " + "[0, parameters.solver.chebIts]."
        )
    N = velocities.N[k] - 1
    ws = np.empty((N, velocities.Ws.shape[3]), dtype=np.float64)
    ws[:N, :] = velocities.Ws[:N, 1, k, :]
    return ws


# @cc.export('lower_dWsdz', float64[:, ::1](VelocityContainer_type, int64))
@jit(
    float64[:, ::1](VelocityContainer_type, int64),
    nopython=True,
    cache=True,
    fastmath=True,
)
def lower_dWsdz(velocities, k):
    if k < 0 or k >= velocities.N.size:
        raise ValueError(
            "In lower_dWsdz, k must be in the range "
            + "[0, parameters.solver.chebIts]."
        )
    N = velocities.N[k]
    dws = np.empty((N, velocities.dWs.shape[3]), dtype=np.float64)
    dws[:N, :] = velocities.dWs[:N, 0, k, :]
    return dws


# @cc.export('upper_dWsdz', float64[:, ::1](VelocityContainer_type, int64))
@jit(
    float64[:, ::1](VelocityContainer_type, int64),
    nopython=True,
    cache=True,
    fastmath=True,
)
def upper_dWsdz(velocities, k):
    if k < 0 or k >= velocities.N.size:
        raise ValueError(
            "In upper_dWsdz, k must be in the range "
            + "[0, parameters.solver.chebIts]."
        )
    N = velocities.N[k] - 1
    dws = np.empty((N, velocities.dWs.shape[3]), dtype=np.float64)
    dws[:N, :] = velocities.dWs[:N, 1, k, :]
    return dws


# @cc.export('Ws_surface', float64[::1](VelocityContainer_type, int64))
@jit(
    float64[::1](VelocityContainer_type, int64),
    nopython=True,
    cache=True,
    fastmath=True,
)
def Ws_surface(velocities, k):
    if k < 0 or k >= velocities.N.size:
        raise ValueError(
            "In lower_Ws, k must be in the range " + "[0, parameters.solver.chebIts]."
        )
    Nbins = velocities.Ws.shape[3]
    ws = np.empty((Nbins), dtype=np.float64)
    ws[:Nbins] = velocities.Ws[0, 0, k, :]
    return ws


# @cc.export('Source_z_dimless', float64[::1](float64[::1], float64))
@jit(float64[::1](float64[::1], float64), nopython=True, cache=True, fastmath=True)
def Source_z_dimless(z, Suzuki_k):
    k = Suzuki_k

    fz = k * k * (1 - z) * np.exp(-k * (1 - z)) / (1 - (1 + k) * np.exp(-k))
    fz[z > 1] = 0

    return fz


@jit(float64(float64, float64), nopython=True, cache=True, fastmath=True)
def Source_z_dimless_value(z, Suzuki_k):
    k = Suzuki_k

    if z > 1:
        fz = 0.0
    else:
        fz = k * k * (1 - z) * np.exp(-k * (1 - z)) / (1 - (1 + k) * np.exp(-k))

    return fz


# @cc.export('Source_z', float64[::1](float64[::1], float64, float64))
@jit(
    float64[::1](float64[::1], float64, float64),
    nopython=True,
    cache=True,
    fastmath=True,
)
def Source_z(z, PlumeH, Suzuki_k):
    xi = z / PlumeH
    k = Suzuki_k

    fz = k * k * (1 - xi) * np.exp(-k * (1 - xi)) / (1 - (1 + k) * np.exp(-k))
    fz[z > PlumeH] = 0

    fz /= PlumeH
    return fz


@jit(
    float64(float64, float64, float64),
    nopython=True,
    cache=True,
    fastmath=True,
)
def Source_z_value(z, PlumeH, Suzuki_k):
    if z > PlumeH:
        fz = 0.0
    else:
        xi = z / PlumeH
        k = Suzuki_k

        fz = k * k * (1 - xi) * np.exp(-k * (1 - xi)) / (1 - (1 + k) * np.exp(-k))

        fz /= PlumeH
    return fz


# if __name__ == "__main__":
#     cc.compile()



from .params import (Parameters, Parameters_type,
                     copy_parameters,
                     update_parameters,
                     save_parameters,
                     load_parameters)

from .solver_params import (SolverParameters, SolverParameters_type)
from .grain_params import (GrainParameters, GrainParameters_type)
from .source_params import (SourceParameters, SourceParameters_type)
from .physical_params import (PhysicalParameters, PhysicalParameters_type)
from .met_params import (MetParameters, MetParameters_type)
from .model_params import (ModelParameters, ModelParameters_type)
from .output_params import (OutputParameters, OutputParameters_type)



# try:
#     from .utilities import *
# except:
#     pass

from .geo import latlon_point_to_utm_code, webmerc, wgs84
from .number_utils import (nice_round_up, nice_round_down, log_levels, lin_levels, log_steps)
from .datetime_strings import string_to_datetime, is_valid_datetime
from .plotting import (plot_rowscols, pad_window)



# from numba.pycc import CC
import numpy as np
from numba import float64, njit

# utilities_cc = CC('utilities')
# utilities_cc.verbose = True


# @utilities_cc.export('interp_ex', 'float64[:](float64[:], float64[:], float64[:])')
@njit(float64[:](float64[:], float64[:], float64[:]))
def interp_ex_array(x, xp, fp):
    i = np.zeros(np.shape(x), dtype=np.int64)
    i[x < xp[0]] = -1
    i[x > xp[-1]] = 1
    y = np.zeros_like(x)
    y[i == -1] = fp[0] + (x[i == -1] - xp[0]) * (fp[1] - fp[0]) / (xp[1] - xp[0])
    y[i == 1] = fp[-1]
    y[i == 0] = np.interp(x[i == 0], xp, fp)
    return y


@njit(float64(float64, float64[:], float64[:]))
def interp_ex_value(x, xp, fp):
    if x < xp[0]:
        y = fp[0] + (x - xp[0]) * (fp[1] - fp[0]) / (xp[1] - xp[0])
    elif x > xp[-1]:
        y = fp[-1]
    else:
        y = np.interp(x, xp, fp)
    return y


# Standard atmosphere functions
# @utilities_cc.export('SA_Temperature', 'float64[::1](float64[::1], float64, float64, float64, float64, float64)')
@njit(float64[::1](float64[::1], float64, float64, float64, float64, float64))
def SA_Temperature_array(z, Ta0, mu, omega, Ht, Hs):
    Ta = np.empty_like(z, dtype=np.float64)
    Ta[z <= Ht] = Ta0 - mu * z[z <= Ht]
    Ta[(Ht < z) * (z < Hs)] = Ta0 - mu * Ht
    Ta[z >= Hs] = Ta0 - mu * Ht + omega * (z[z >= Hs] - Hs)
    return Ta


@njit(float64(float64, float64, float64, float64, float64, float64))
def SA_Temperature_value(z, Ta0, mu, omega, Ht, Hs):
    if z <= Ht:
        Ta = Ta0 - mu * z
    elif z >= Hs:
        Ta = Ta0 - mu * Ht + omega * (z - Hs)
    else:
        Ta = Ta0 - mu * Ht
    return Ta


# @utilities_cc.export('SA_Pressure', 'float64[::1](float64[::1], float64, float64, float64, float64, float64, float64, float64, float64)')
@njit(
    float64[::1](
        float64[::1],
        float64,
        float64,
        float64,
        float64,
        float64,
        float64,
        float64,
        float64,
    )
)
def SA_Pressure_array(z, Ta0, Pa0, mu, omega, Ht, Hs, g, Ra):
    P0 = Pa0 / np.power(Ta0, g / Ra / mu)

    a = g / Ra / mu
    b = g / Ra / omega

    Pa = np.empty_like(z, dtype=np.float64)

    Pa[z <= Ht] = P0 * np.power(Ta0 - mu * z[z <= Ht], a)
    Pa[(Ht < z) * (z < Hs)] = (
        P0
        * np.power(Ta0 - mu * Ht, a)
        * np.exp(-g * (z[(Ht < z) * (z < Hs)] - Ht) / Ra / (Ta0 - mu * Ht))
    )
    Pa[z >= Hs] = (
        P0
        * np.power(Ta0 - mu * Ht, (a + b))
        * np.exp(-g * (Hs - Ht) / Ra / (Ta0 - mu * Ht))
        * np.power(Ta0 - mu * Ht + omega * (z[z >= Hs] - Hs), -b)
    )
    return Pa


@njit(
    float64(
        float64,
        float64,
        float64,
        float64,
        float64,
        float64,
        float64,
        float64,
        float64,
    )
)
def SA_Pressure_value(z, Ta0, Pa0, mu, omega, Ht, Hs, g, Ra):
    P0 = Pa0 / np.power(Ta0, g / Ra / mu)

    a = g / Ra / mu
    b = g / Ra / omega

    if z <= Ht:
        Pa = P0 * np.power(Ta0 - mu * z, a)
    elif z >= Hs:
        Pa = (
            P0
            * np.power(Ta0 - mu * Ht, (a + b))
            * np.exp(-g * (Hs - Ht) / Ra / (Ta0 - mu * Ht))
            * np.power(Ta0 - mu * Ht + omega * (z - Hs), -b)
        )
    else:
        Pa = (
            P0
            * np.power(Ta0 - mu * Ht, a)
            * np.exp(-g * (z - Ht) / Ra / (Ta0 - mu * Ht))
        )

    return Pa


# @utilities_cc.export('SA_Density', 'float64[::1](float64[::1], float64, float64, float64, float64, float64, float64, float64, float64)')
@njit(
    float64[::1](
        float64[::1],
        float64,
        float64,
        float64,
        float64,
        float64,
        float64,
        float64,
        float64,
    )
)
def SA_Density_array(z, Ta0, Pa0, mu, omega, Ht, Hs, g, Ra):

    P0 = Pa0 / np.power(Ta0, g / Ra / mu)

    a = g / Ra / mu
    b = g / Ra / omega

    Pa = np.empty_like(z, dtype=np.float64)

    Pa[z <= Ht] = P0 * np.power(Ta0 - mu * z[z <= Ht], a)
    Pa[(Ht < z) * (z < Hs)] = (
        P0
        * np.power(Ta0 - mu * Ht, a)
        * np.exp(-g * (z[(Ht < z) * (z < Hs)] - Ht) / Ra / (Ta0 - mu * Ht))
    )
    Pa[z >= Hs] = (
        P0
        * np.power(Ta0 - mu * Ht, (a + b))
        * np.exp(-g * (Hs - Ht) / Ra / (Ta0 - mu * Ht))
        * np.power(Ta0 - mu * Ht + omega * (z[z >= Hs] - Hs), -b)
    )

    Ta = np.empty_like(z, dtype=np.float64)
    Ta[z <= Ht] = Ta0 - mu * z[z <= Ht]
    Ta[(Ht < z) * (z < Hs)] = Ta0 - mu * Ht
    Ta[z >= Hs] = Ta0 - mu * Ht + omega * (z[z >= Hs] - Hs)

    rhoa = Pa / (Ra * Ta)
    return rhoa


@njit(
    float64(
        float64,
        float64,
        float64,
        float64,
        float64,
        float64,
        float64,
        float64,
        float64,
    )
)
def SA_Density_value(z, Ta0, Pa0, mu, omega, Ht, Hs, g, Ra):

    P0 = Pa0 / np.power(Ta0, g / Ra / mu)

    a = g / Ra / mu
    b = g / Ra / omega

    Pa = SA_Pressure_value(z, Ta0, Pa0, mu, omega, Ht, Hs, g, Ra)

    Ta = SA_Temperature_value(z, Ta0, mu, omega, Ht, Hs)

    rhoa = Pa / (Ra * Ta)
    return rhoa


# if __name__ == "__main__":
# utilities_cc.compile()



from collections import OrderedDict
from numba.experimental import jitclass
from numba import float64


met_spec = OrderedDict()
met_spec['U_scale'] = float64      # scale for wind velocity
met_spec['Ws_scale'] = float64[:]  # settling scale for each grain class


@jitclass(met_spec)
class MetParameters():
    def __init__(self, U_scale, Ws_scale):
        self.U_scale = U_scale
        self.Ws_scale = Ws_scale


# pylint: disable=E1101
MetParameters_type = MetParameters.class_type.instance_type



import os
from datetime import datetime

import numpy as np
import pandas as pd
import requests
import xarray as xr


class gfs_archive:
    def __init__(self, run_datetime: str, fxx: int):

        self.fxx = fxx
        self.date = pd.to_datetime(run_datetime)

        self.url = self._aws_site()
        self.idx_url = self.url + ".idx"

        grib_exists = self._check_grib()
        idx_exists = self._check_idx()

        if grib_exists and idx_exists:
            self.idx = self._get_idx_as_dataframe()

            self.levels = self.idx.loc[
                (self.idx["level"].str.match("(\d+(?:\.\d+)?) mb"))
            ].level.unique()

    def _aws_site(self):
        return f"https://noaa-gfs-bdp-pds.s3.amazonaws.com/gfs.{self.date:%Y%m%d/%H}/atmos/gfs.t{self.date:%H}z.pgrb2.0p25.f{self.fxx:03d}"

    def _check_grib(self):
        head = requests.head(url)
        check_exists = head.ok
        if check_exists:
            check_content = int(head.raw.info()["Content-Length"]) > 1_000_000
            return check_exists and check_content
        else:
            return False

    def _check_idx(self):
        idx_exists = requests.head(self.idx_url).ok
        return idx_exists

    def _get_idx_as_dataframe(self):
        df = pd.read_csv(
            self.idx_url,
            sep=":",
            names=[
                "grib_message",
                "start_byte",
                "reference_time",
                "variable",
                "level",
                "forecast_time",
                "?",
                "??",
                "???",
            ],
        )

        # Format the DataFrame
        df["reference_time"] = pd.to_datetime(df.reference_time, format="d=%Y%m%d%H")
        df["valid_time"] = df["reference_time"] + pd.to_timedelta(f"{self.fxx}H")
        df["start_byte"] = df["start_byte"].astype(int)
        df["end_byte"] = df["start_byte"].shift(-1, fill_value="")
        # TODO: Check this works: Assign the ending byte for the last row...
        # TODO: df["end_byte"] = df["start_byte"].shift(-1, fill_value=requests.get(self.grib, stream=True).headers['Content-Length'])
        # TODO: Based on what Karl Schnieder did.
        df["range"] = df.start_byte.astype(str) + "-" + df.end_byte.astype(str)
        df = df.reindex(
            columns=[
                "grib_message",
                "start_byte",
                "end_byte",
                "range",
                "reference_time",
                "valid_time",
                "variable",
                "level",
                "forecast_time",
                "?",
                "??",
                "???",
            ]
        )

        df = df.dropna(how="all", axis=1)
        df = df.fillna("")

        df["search_this"] = (
            df.loc[:, "variable":]
            .astype(str)
            .apply(
                lambda x: ":" + ":".join(x).rstrip(":").replace(":nan:", ":"),
                axis=1,
            )
        )

        # Attach some attributes
        df.attrs = dict(
            url=self.idx_url,
            description="Inventory index file for the GRIB2 file.",
            lead_time=self.fxx,
            datetime=self.date,
        )

        return df

    def read_idx(self, searchString=None):
        """
        Inspect the GRIB2 file contents by reading the index file.

        This reads index files created with the wgrib2 utility.

        Parameters
        ----------
        searchString : str
            Filter dataframe by a searchString regular expression.
            Searches for strings in the index file lines, specifically
            the variable, level, and forecast_time columns.
            Execute ``_searchString_help()`` for examples of a good
            searchString.

            .. include:: ../../user_guide/searchString.rst

        Returns
        -------
        A Pandas DataFrame of the index file.
        """

        # Filter DataFrame by searchString
        if searchString not in [None, ":"]:
            logic = self.idx.search_this.str.contains(searchString)
            if logic.sum() == 0:
                print(
                    f"No GRIB messages found. There might be something wrong with {searchString=}"
                )
            df = self.idx.loc[logic]
        return df

    def download_grib(self, searchString, outFile="./gfs_grib_file.grib2"):

        grib_source = self.url

        # Download subsets of the file by byte range with cURL.
        # > Instead of using a single curl command for each row,
        # > group adjacent messages in the same curl command.

        # Find index groupings
        # TODO: Improve this for readability
        # https://stackoverflow.com/a/32199363/2383070
        idx_df = self.read_idx(searchString)
        li = idx_df.index
        inds = (
            [0]
            + [ind for ind, (i, j) in enumerate(zip(li, li[1:]), 1) if j - i > 1]
            + [len(li) + 1]
        )

        curl_groups = [li[i:j] for i, j in zip(inds, inds[1:])]
        curl_ranges = []
        group_dfs = []
        for i, group in enumerate(curl_groups):
            _df = idx_df.loc[group]
            curl_ranges.append(f"{_df.iloc[0].start_byte}-{_df.iloc[-1].end_byte}")
            group_dfs.append(_df)

            for i, (range, _df) in enumerate(zip(curl_ranges, group_dfs)):

                if i == 0:
                    # If we are working on the first item, overwrite the existing file...
                    curl = f"curl -s --range {range} {grib_source} > {outFile}"
                else:
                    # ...all other messages are appended to the subset file.
                    curl = f"curl -s --range {range} {grib_source} >> {outFile}"
                os.system(curl)

    def profiles(self, latitude, longitude):

        data_P = np.zeros(levels.size)
        data_Z = np.zeros(levels.size)
        data_T = np.zeros(levels.size)
        data_U = np.zeros(levels.size)
        data_V = np.zeros(levels.size)

        outFile = "./gfs_grib_file.grib2"

        for j, l in enumerate(self.levels):
            data_P[j] = np.float64(l.replace(" mb", "")) * 100

            self.download_grib(f":HGT:{l}", outFile=outFile)

            Z = xr.load_dataset(outFile, engine="cfgrib")
            data_Z[j] = np.float64(
                Z["gh"]
                .interp(latitude=latitude, longitude=longitude, method="cubic")
                .values
            )
            self.download_grib(f":TMP:{l}", outFile="./gfs_grib_file.grib2")
            T = xr.load_dataset(outFile, engine="cfgrib")
            data_T[j] = np.float64(
                T["t"]
                .interp(latitude=latitude, longitude=longitude, method="cubic")
                .values
            )
            self.download_grib(f":(?:U|V)GRD:{l}", outFile="./gfs_grib_file.grib2")
            UV = xr.load_dataset(outFile, engine="cfgrib")
            uv_interp = UV.interp(
                latitude=latitude, longitude=longitude, method="cubic"
            )
            data_U[j] = np.float64(uv_interp["u"].values)
            data_V[j] = np.float64(uv_interp["v"].values)

        df = pd.DataFrame(columns=["Z", "T", "P", "U", "V"])
        df["Z"] = data_Z
        df["T"] = data_T
        df["P"] = data_P
        df["U"] = data_U
        df["V"] = data_V

        df = df.dropna()
        df = df.sort_values("Z", ignore_index=True)

        return df



import datetime
import importlib.resources as pkg_resources
import re
from os.path import isfile

import pandas as pd
from numpy import isfinite
from prompt_toolkit import HTML, print_formatted_text, prompt
from prompt_toolkit.completion import PathCompleter, WordCompleter
from prompt_toolkit.styles import Style
from prompt_toolkit.validation import ValidationError, Validator

from ashdisperse import data
from ashdisperse.utilities import is_valid_datetime, string_to_datetime


def isnotebook():
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            ret = True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            ret = False  # Terminal running IPython
        else:
            ret = False  # Other type (?)
        return ret
    except NameError:
        return False  # Probably standard Python interpreter


__ISNOTEBOOK = isnotebook()

style = Style.from_dict(
    {
        "title": "#3b76fe bold",
        "text": "#3b76fe",
        "query": "#33cc33",
        "query_ans": "#33cc33 italic",
        "error": "#ff0000",
        "fatal_error": "#ff0000 bold",
    }
)

gvpList = pkg_resources.open_text(data, "GVP_Volcano_List.csv")

gvp_list = pd.read_csv(gvpList)


def print_title(title_text):
    if __ISNOTEBOOK:
        print(title_text)
    else:
        print_formatted_text(
            HTML("<title>{}\n</title>".format(title_text)), style=style
        )


def print_text(text):
    if __ISNOTEBOOK:
        print(text)
    else:
        print_formatted_text(HTML("<text> {} </text>".format(text)), style=style)


def query_yes_no(question, default="yes"):

    if default is None:
        ans = " [y/n] "
    elif default == "yes":
        ans = " [Y/n] "
    elif default == "no":
        ans = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}

    if __ISNOTEBOOK:
        question_text = "{} {}:".format(question, ans)
        error_text = "Please respond with 'yes' or 'no' (or 'y' or 'n')"
    else:
        question_text = HTML(
            "<query> {} </query> <query_ans>".format(question)
            + "{}: </query_ans>".format(ans)
        )
        error_text = HTML(
            "<error> Please respond with 'yes' or 'no'" + "(or 'y' or 'n').\n </error>"
        )

    while True:
        if __ISNOTEBOOK:
            choice = input(question_text).lower()
        else:
            choice = prompt(question_text, style=style).lower()

        if default is not None and choice == "":
            return valid[default]
        if choice in valid:
            return valid[choice]

        if __ISNOTEBOOK:
            print(error_text)
        else:
            print_formatted_text(error_text, style=style)


def query_choices(question, choices=["y", "n"], default=None):
    """Asks a question via input() with answer in choices.

    Args:
        question: a string that is presented to the user.
        choices: a list of allowed responses
        **default: Optional; the presumed answer if the user just hits <Enter>.

    Returns:
        The "answer" return value is True for "yes" or False for "no".
    """
    ans = "[" + "/".join(choices) + "]"
    if default is not None:
        ans += " (default {})".format(default)

    if __ISNOTEBOOK:
        question_text = "{} {}: ".format(question, ans)
        error_text = "Please respond with one of " + ",".join(choices) + "."
    else:
        question_text = HTML(
            "<query> {} </query> <query_ans>".format(question)
            + "{}: </query_ans>".format(ans)
        )
        error_text = HTML(
            "<error> Please respond with one of " + ",".join(choices) + ".\n </error>"
        )
        ans_completer = WordCompleter(choices)

    while True:
        if __ISNOTEBOOK:
            choice = input(question_text)
        else:
            choice = prompt(
                question_text,
                style=style,
                completer=ans_completer,
                complete_while_typing=True,
            )

        if default is not None and choice == "":
            return default
        if choice in choices:
            return choice
        if __ISNOTEBOOK:
            print(error_text)
        else:
            print_formatted_text(error_text, style=style)


def query_change_value(question, default=None, answer_type=float):
    """Ask a question via input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be a value
    "answer_type" is the type of the answer expected.

    """

    if default is not None and not isinstance(default, answer_type):
        raise ValueError(
            "In query_change_value, default value {}".format(default)
            + "must be of type {}".format(answer_type)
        )

    if default is None:
        ans = " "
    else:
        ans = " [return for default = {}] ".format(default)

    def is_valid_num(text):
        if text == "":
            return True
        try:
            answer_type(text)
            return True
        except ValueError:
            return False

    if __ISNOTEBOOK:
        question_text = "{} {}:".format(question, ans)
    else:
        validator = Validator.from_callable(
            is_valid_num,
            error_message="This input contains non-numeric characters",
            move_cursor_to_end=True,
        )
        question_text = HTML(
            "<query> {} </query> <query_ans>".format(question)
            + " {}: </query_ans>".format(ans)
        )

    while True:
        if __ISNOTEBOOK:
            choice = input(question_text)
        else:
            choice = prompt(question_text, style=style, validator=validator)

        if default is not None and choice == "":
            ret = default
        else:
            ret = answer_type(choice)
        return ret


def query_set_value(question, answer_type=float):
    """Ask a question via input() and return their answer.

    "question" is a string that is presented to the user.
    "answer_type" is the type of the answer expected.

    """

    def is_valid_num(text):
        try:
            answer_type(text)
            return True
        except ValueError:
            return False

    if __ISNOTEBOOK:
        question_text = "{} : ".format(question)
    else:
        validator = Validator.from_callable(
            is_valid_num,
            error_message="This input contains non-numeric characters",
            move_cursor_to_end=True,
        )
        question_text = HTML("<query> {} </query> : ".format(question))

    while True:
        if __ISNOTEBOOK:
            choice = input(question_text)
        else:
            choice = prompt(question_text, style=style, validator=validator)

        return answer_type(choice)


def query_datetime(question):
    if __ISNOTEBOOK:
        question_text = "{} :".format(question)
    else:
        validator = Validator.from_callable(
            is_valid_datetime,
            error_message="This input is not recognized as a datetime string",
            move_cursor_to_end=True,
        )
        question_text = HTML("<query> {} </query> : ".format(question))

    while True:
        if __ISNOTEBOOK:
            input_datetime = input(question_text)
        else:
            input_datetime = prompt(question_text, style=style, validator=validator)

        output_datetime = string_to_datetime(input_datetime)

        return output_datetime


def is_valid_float(num):
    try:
        float(num)
        return True
    except ValueError:
        return False


def is_valid_lat(lat):
    lat = float(lat)
    if lat > 90 or lat < -90:
        ret = False
    else:
        ret = True
    return ret


def is_valid_lon(lon):
    lon = float(lon)
    if lon > 360 or lon < -180:
        ret = False
    else:
        ret = True
    return ret


def is_valid_latlon(latlon):
    try:
        lat, lon = latlon.split(",", 2)
        lat = float(lat)
        lon = float(lon)
        if is_valid_lat(lat) & is_valid_lon(lon):
            ret = (lat, lon)
        else:
            ret = (None, None)
        return ret
    except ValueError:
        return (None, None)


def has_comma(text):
    text_split = text.split(",")
    if len(text_split) != 2:
        ret = False
    else:
        ret = True
    return ret


def get_latlon_from_gvp_name(name):
    this_volc = gvp_list.loc[gvp_list.Volcano_Name == name]
    if this_volc["Primary_Volcano_Type"].str.contains("Synonym").values[0]:
        name = this_volc["Primary_Volcano_Type"].values[0].split("of ")[1]
        this_volc = gvp_list.loc[gvp_list.Volcano_Name == name]

    lat = this_volc["Latitude"].values[0]
    lon = this_volc["Longitude"].values[0]

    if not isfinite(lat) and not isfinite(lon):
        entries = gvp_list.loc[
            gvp_list.Volcano_Number == this_volc["Volcano_Number"].values[0]
        ]
        volc = entries.loc[isfinite(entries.Latitude)]
        name = volc["Volcano_Name"].values[0]
        lat = volc["Latitude"].values[0]
        lon = volc["Longitude"].values[0]

    return name, lat, lon


class NameLatLonValidator(Validator):
    def validate(self, document):
        text = document.text

        # if not has_comma(text):

        if has_comma(text) and (
            text[0].isnumeric() or (text[0] == "-" and text[1].isnumeric())
        ):
            lat_i = text.find(",")
            lat, lon = text.split(",", 2)
            if not is_valid_float(lat):
                raise ValidationError(
                    message="Specify latitude in decimal degrees", cursor_position=lat_i
                )
            if not is_valid_lat(lat):
                raise ValidationError(
                    message="Specify latitude in decimal degrees in "
                    + "range -90 -- 90",
                    cursor_position=lat_i,
                )
            if not is_valid_float(lon):
                raise ValidationError(
                    message="Specify longitude in decimal degrees",
                    cursor_position=len(text),
                )
            if not is_valid_lon(lon):
                raise ValidationError(
                    message="Specify latitude in decimal degrees "
                    + "in range -180--180 or 0--360",
                    cursor_position=len(text),
                )
        else:
            if text not in gvp_list.Volcano_Name.values:
                raise ValidationError(
                    message="Specify location as volcano name"
                    + " or as latitude and longitude "
                    + "(as decimal degrees in format lat, lon)",
                    cursor_position=len(text),
                )


def query_latlon():
    latlon_query_text = (
        "Enter Volcano name or give latitude and "
        + "longitude of source (as decimal degrees in format lat, lon): "
    )

    if __ISNOTEBOOK:
        while True:
            latlon = input(latlon_query_text)
            lat, lon = is_valid_latlon(latlon)
            if lat is not None and lon is not None:
                name = None
                break

            if latlon in gvp_list.Volcano_Name.values:
                name, lat, lon = get_latlon_from_gvp_name(latlon)
                break

            print(
                "Not a recognized volcano name nor a valid "
                + "latitude-longitude coordinate"
            )
            print(
                "Specify latitude and longitude "
                + "(as decimal degrees in format lat, lon)"
            )
    else:
        latlon_query = HTML("<query> {} </query>".format(latlon_query_text))
        name_cmplter = WordCompleter(gvp_list.Volcano_Name.unique())
        latlon = prompt(
            latlon_query,
            completer=name_cmplter,
            validator=NameLatLonValidator(),
            style=style,
        )
        if has_comma(latlon) and latlon[0].isnumeric():
            lat, lon = is_valid_latlon(latlon)
            name = None
        else:
            name, lat, lon = get_latlon_from_gvp_name(latlon)

    return name, lat, lon


def query_met_file():
    met_query = "Path of met file (netCDF): "
    if __ISNOTEBOOK:
        while True:
            met_file = input(met_query)
            if not isfile(met_file):
                print_text("Met file {} does not exist".format(met_file))
            else:
                break
    else:
        met_file = prompt(met_query, completer=PathCompleter(expanduser=True))
    return met_file



from branca.element import MacroElement

from jinja2 import Template

class BindColormap(MacroElement):
    """Binds a colormap to a given layer.

    Parameters
    ----------
    colormap : branca.colormap.ColorMap
        The colormap to bind.
    """
    def __init__(self, layer, colormap):
        super(BindColormap, self).__init__()
        self.layer = layer
        self.colormap = colormap
        self._template = Template(u"""
        {% macro script(this, kwargs) %}
            {{this.colormap.get_name()}}.svg[0][0].style.display = 'block';
            {{this._parent.get_name()}}.on('overlayadd', function (eventLayer) {
                if (eventLayer.layer == {{this.layer.get_name()}}) {
                    {{this.colormap.get_name()}}.svg[0][0].style.display = 'block';
                }});
            {{this._parent.get_name()}}.on('overlayremove', function (eventLayer) {
                if (eventLayer.layer == {{this.layer.get_name()}}) {
                    {{this.colormap.get_name()}}.svg[0][0].style.display = 'none';
                }});
        {% endmacro %}
        """)  # noqa


from math import ceil, floor
import numpy as np


def nice_round_up(val, mag=None):
    """Rounds up a number to a nice value of the form a*10^b for integer
        a and b.

    Args:
      val: the number to round.
      **mag: Optional; the value of the exponent b.

    Returns:
      A number rounded.
    """
    if mag is None:
        mag = 10**int(np.log10(val))
    return ceil(val/mag)*mag


def nice_round_down(val, mag=None):
    """Rounds down a number to a nice value of the form a*10^b for integer
        a and b.

    Args:
      val: the number to round.
      **mag: Optional; the value of the exponent b.

    Returns:
      A number rounded.
    """
    if mag is None:
        mag = 10**int(np.log10(val))
    return floor(val/mag)*mag


def log_levels(vmin, vmax):
    mag_low = floor(np.log10(vmin))
    mag_high = floor(np.log10(vmax))

    vals = np.arange(1, 10)
    levels = []
    for mag in range(mag_low, mag_high+1):
        levels.append(vals*10**mag)

    levels = np.asarray(levels).flatten()

    return levels[(levels >= vmin) & (levels <= vmax)]

def log_steps(vmin, vmax, step=10, include_max=True):
    if vmin<=0:
        raise ValueError('vmin must be positive, received {}'.format(vmin))
    levels = []
    this_level = vmin
    while this_level<vmax:
        levels.append(this_level)
        this_level *= step
    
    if include_max:
        if vmax not in levels:
            levels.append(vmax)

    return levels

def lin_levels(vmin, vmax, num=10):
    return np.linspace(vmin, vmax, num)



import numpy as np
from numba import complex128, float64, int64, jit, njit

# from numba.pycc import CC
from numba.types import Tuple

from ..containers.cheb import ChebContainer_type
from ..containers.velocities import VelocityContainer_type
from ..met.met import MetData_type
from ..params.params import Parameters_type
from ..spectral.cheb import convergedCoeffs
from .getters import (
    Source_z_dimless,
    lower_dWsdz,
    lower_U,
    lower_V,
    lower_Ws,
    upper_dWsdz,
    upper_U,
    upper_V,
    upper_Ws,
)
from .rk import integrateTo

# pylint: disable=C0103, E501

# cc = CC('core')
# cc.verbose = True


# LowerODE
# @cc.export('LowerODE', Tuple((complex128[::1], complex128[::1], complex128[::1]))(float64, float64, complex128, int64, ChebContainer_type, Parameters_type, VelocityContainer_type))
@jit(
    Tuple((complex128[::1], complex128[::1], complex128[::1]))(
        float64,
        float64,
        complex128,
        int64,
        ChebContainer_type,
        Parameters_type,
        VelocityContainer_type,
    ),
    nopython=True,
    cache=True,
    fastmath=True,
)
def LowerODE(kx, ky, fxy_ij, grain_i, cheby, parameters, velocities):
    pi = np.pi
    Lx = parameters.model.Lx[grain_i]
    Ly = parameters.model.Ly[grain_i]
    Pe = parameters.model.Peclet_number
    Vratio = parameters.model.Velocity_ratio[grain_i]
    R = parameters.model.Diffusion_ratio

    for k, N in enumerate(cheby.N):
        x, Tn, dTn, d2Tn = cheby.get_cheb(k)
        U = lower_U(velocities, k)
        V = lower_V(velocities, k)
        Ws = lower_Ws(velocities, k)[:, grain_i]
        dWsdz = lower_dWsdz(velocities, k)[:, grain_i]

        c2 = 4.0 * Pe * Vratio / R * np.ones_like(x, dtype=np.complex128)
        c2a = np.reshape(np.repeat(c2, N), (N, N))
        L2 = np.multiply(c2a, d2Tn)

        c1 = 2 * Ws
        c1a = np.reshape(np.repeat(c1, N), (N, N))
        L1 = np.multiply(c1a, dTn)

        c0 = (
            -1j * pi * (kx / Lx * U + ky / Ly * V)
            + dWsdz
            - pi * pi * Pe / Vratio * ((kx / Lx) ** 2 + (ky / Ly) ** 2)
        )
        c0a = np.reshape(np.repeat(c0, N), (N, N))
        L0 = np.multiply(c0a, Tn)

        L = L2 + L1 + L0

        b = np.zeros((N, 3), dtype=np.complex128)
        fz = Source_z_dimless(0.5 * (x + 1), parameters.source.Suzuki_k)

        L[0, :] = Tn[0, :]
        L[-1, :] = Tn[-1, :]

        b[:, 0] = -fxy_ij * fz
        b[0, 0] = 0.0
        b[-1, 0] = 0.0

        b[0, 1] = 1.0
        b[-1, 1] = 0.0

        b[0, 2] = 0.0
        b[-1, 2] = 1.0

        coeffs = np.linalg.solve(L, b)  # .astype(np.complex128)

        co0 = convergedCoeffs(coeffs[:, 0].flatten(), parameters.solver.epsilon)
        co1 = convergedCoeffs(coeffs[:, 1].flatten(), parameters.solver.epsilon)
        co2 = convergedCoeffs(coeffs[:, 2].flatten(), parameters.solver.epsilon)

        if (co0.size < N) and (co1.size < N) and (co2.size < N):
            break

    return co0, co1, co2


# LowerODE
# @cc.export('LowerODE', Tuple((complex128[::1], complex128[::1], complex128[::1]))(float64, float64, complex128, int64, ChebContainer_type, Parameters_type, VelocityContainer_type))
@jit(
    complex128[::1](
        float64,
        float64,
        complex128,
        int64,
        ChebContainer_type,
        Parameters_type,
        VelocityContainer_type,
    ),
    nopython=True,
    cache=True,
    fastmath=True,
)
def LowerODE_f(kx, ky, fxy_ij, grain_i, cheby, parameters, velocities):
    pi = np.pi
    Lx = parameters.model.Lx[grain_i]
    Ly = parameters.model.Ly[grain_i]
    Pe = parameters.model.Peclet_number
    Vratio = parameters.model.Velocity_ratio[grain_i]
    R = parameters.model.Diffusion_ratio

    for k, N in enumerate(cheby.N):
        x, Tn, dTn, d2Tn = cheby.get_cheb(k)
        U = lower_U(velocities, k)
        V = lower_V(velocities, k)
        Ws = lower_Ws(velocities, k)[:, grain_i]
        dWsdz = lower_dWsdz(velocities, k)[:, grain_i]

        c2 = 4.0 * Pe * Vratio / R * np.ones_like(x, dtype=np.complex128)
        c2a = np.reshape(np.repeat(c2, N), (N, N))
        L2 = np.multiply(c2a, d2Tn)

        c1 = 2 * Ws
        c1a = np.reshape(np.repeat(c1, N), (N, N))
        L1 = np.multiply(c1a, dTn)

        c0 = (
            -1j * pi * (kx / Lx * U + ky / Ly * V)
            + dWsdz
            - pi * pi * Pe / Vratio * ((kx / Lx) ** 2 + (ky / Ly) ** 2)
        )
        c0a = np.reshape(np.repeat(c0, N), (N, N))
        L0 = np.multiply(c0a, Tn)

        L = L2 + L1 + L0

        b = np.zeros((N, 1), dtype=np.complex128)
        fz = Source_z_dimless(0.5 * (x + 1), parameters.source.Suzuki_k)

        L[0, :] = dTn[0, :]
        L[-1, :] = Tn[-1, :]

        b[:, 0] = -fxy_ij * fz
        b[0, 0] = 0.0
        b[-1, 0] = 0.0

        coeffs = np.linalg.solve(L, b)  # .astype(np.complex128)

        co0 = convergedCoeffs(coeffs[:, 0].flatten(), parameters.solver.epsilon)

        if co0.size < N:
            break

    return co0


@jit(
    complex128[::1](
        float64,
        float64,
        int64,
        ChebContainer_type,
        Parameters_type,
        VelocityContainer_type,
        int64,
        complex128,
    ),
    nopython=True,
    cache=True,
    fastmath=True,
)
def LowerODE_0(kx, ky, grain_i, cheby, parameters, velocities, b_indx, b_m):
    pi = np.pi
    Lx = parameters.model.Lx[grain_i]
    Ly = parameters.model.Ly[grain_i]
    Pe = parameters.model.Peclet_number
    Vratio = parameters.model.Velocity_ratio[grain_i]
    R = parameters.model.Diffusion_ratio

    for k, N in enumerate(cheby.N):
        x, Tn, dTn, d2Tn = cheby.get_cheb(k)
        U = lower_U(velocities, k)
        V = lower_V(velocities, k)
        Ws = lower_Ws(velocities, k)[:, grain_i]
        dWsdz = lower_dWsdz(velocities, k)[:, grain_i]

        c2 = 4.0 * Pe * Vratio / R * np.ones_like(x, dtype=np.complex128)
        c2a = np.reshape(np.repeat(c2, N), (N, N))
        L2 = np.multiply(c2a, d2Tn)

        c1 = 2 * Ws
        c1a = np.reshape(np.repeat(c1, N), (N, N))
        L1 = np.multiply(c1a, dTn)

        c0 = (
            -1j * pi * (kx / Lx * U + ky / Ly * V)
            + dWsdz
            - pi * pi * Pe / Vratio * ((kx / Lx) ** 2 + (ky / Ly) ** 2)
        )
        c0a = np.reshape(np.repeat(c0, N), (N, N))
        L0 = np.multiply(c0a, Tn)

        L = L2 + L1 + L0

        b = np.zeros((N, 1), dtype=np.complex128)

        L[0, :] = dTn[0, :]
        L[-1, :] = Tn[b_indx, :]

        b[0, 0] = 0.0
        b[-1, 0] = b_m

        coeffs = np.linalg.solve(L, b)  # .astype(np.complex128)

        co = convergedCoeffs(coeffs[:, 0].flatten(), parameters.solver.epsilon)

        if co.size < N:
            break

    return co


# UpperODE
# @cc.export('UpperODE', complex128[::1](float64, float64, int64, ChebContainer_type, Parameters_type, VelocityContainer_type))
@jit(
    complex128[::1](
        float64,
        float64,
        int64,
        ChebContainer_type,
        Parameters_type,
        VelocityContainer_type,
    ),
    nopython=True,
    cache=True,
    fastmath=True,
)
def UpperODE(kx, ky, grain_i, cheby, parameters, velocities):
    pi = np.pi
    Lx = parameters.model.Lx[grain_i]
    Ly = parameters.model.Ly[grain_i]
    Pe = parameters.model.Peclet_number
    Vratio = parameters.model.Velocity_ratio[grain_i]
    R = parameters.model.Diffusion_ratio

    for k, N in enumerate(cheby.N):
        x, Tn, dTn, d2Tn = cheby.get_cheb(k)
        U = upper_U(velocities, k)
        V = upper_V(velocities, k)
        Ws = upper_Ws(velocities, k)[:, grain_i]
        dWsdz = upper_dWsdz(velocities, k)[:, grain_i]

        c2 = np.zeros_like(x, dtype=np.complex128)
        c2[:-1] = ((1 - x[:-1]) ** 4) / 4 * Pe * Vratio / R
        c2a = np.reshape(np.repeat(c2, N), (N, N))
        L2 = np.multiply(c2a, d2Tn)

        c1 = np.zeros_like(x, dtype=np.complex128)
        c1[:-1] = ((1 - x[:-1]) ** 2) / 2 * (Ws - (1 - x[:-1]) * Pe * Vratio / R)
        c1[-1] = 0
        c1a = np.reshape(np.repeat(c1, N), (N, N))
        L1 = np.multiply(c1a, dTn)

        c0 = np.zeros_like(x, dtype=np.complex128)
        c0[:-1] = (
            -1j * pi * (kx / Lx * U + ky / Ly * V)
            + dWsdz
            - pi * pi * Pe / Vratio * ((kx / Lx) ** 2 + (ky / Ly) ** 2)
        )
        c0[-1] = 0
        c0a = np.reshape(np.repeat(c0, N), (N, N))
        L0 = np.multiply(c0a, Tn)

        L = L2 + L1 + L0
        b = np.zeros((N, 1), dtype=np.complex128)

        L[0, :] = Tn[0, :]
        L[-1, :] = Tn[-1, :]

        b[0, 0] = 1.0

        coeffs = np.linalg.solve(L, b)  # .astype(np.complex128)

        co = convergedCoeffs(coeffs[:, 0].flatten(), parameters.solver.epsilon)

        if co.size < N:
            break

    # co = np.ascontiguousarray(new_coeffs[0].flatten())

    return co


@jit(
    complex128[::1](
        float64,
        float64,
        complex128,
        int64,
        Parameters_type,
        MetData_type,
    ),
    nopython=True,
    cache=True,
    fastmath=True,
)
# @njit(cache=True, fastmath=True)
def rkBVP(kx, ky, fxy_ij, grain_i, parameters, Met):

    Nz = parameters.output.Nz

    x0 = 0.0

    y = np.zeros((4), dtype=np.complex128)
    y[0] = 0.0
    y[1] = 0.0
    y[2] = 0.0
    y[3] = 1.0

    dy = np.zeros((4), dtype=np.complex128)
    dy[0] = 0.0
    dy[1] = 0.0
    dy[2] = 1.0
    dy[3] = 0.0

    h = 1e-8

    u = np.zeros((Nz), dtype=np.complex128)
    v = np.zeros((Nz), dtype=np.complex128)
    C = np.zeros((Nz), dtype=np.complex128)

    for j in range(Nz):

        z = parameters.output.altitudes[-j]

        x = np.exp(-z)

        y, dy, h = integrateTo(
            x0, x, y, dy, h, grain_i, kx, ky, fxy_ij, parameters, Met
        )

        u[j] = y[0]
        v[j] = y[1]

        x0 = x

    x = 1.0
    y, dy, h = integrateTo(x0, x, y, dy, h, grain_i, kx, ky, fxy_ij, parameters, Met)

    du1 = y[1]
    dv1 = y[3]

    A = -du1 / dv1
    C = u + A * v

    return C


# if __name__ == "__main__":
#     cc.compile()



from distutils.util import convert_path

import setuptools

# from ashdisperse.utilities.utilities import utilities_cc
# from ashdisperse.spectral.cheb import cc as cheb_cc
# from ashdisperse.core.getters import cc as getters_cc
# from ashdisperse.core.core import cc as core_cc

main_ns = {}
ver_path = convert_path("ashdisperse/version.py")
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

setuptools.setup(
    name="ashdisperse",
    version=main_ns["__version__"],
    author="Mark Woodhouse",
    author_email="mark.woodhouse@bristol.ac.uk",
    description="Steady state advection-diffusion-sedimentation solver for volcanic ash dispersion",
    packages=setuptools.find_packages(),
    include_package_data=True,
    # ext_modules=[
    #     utilities_cc.distutils_extension(),
    #     cheb_cc.distutils_extension(),
    #     getters_cc.distutils_extension(),
    #     core_cc.distutils_extension(),
    # ],
    install_requires=[
        "numpy",
        "numba",
        "intel-openmp;platform_system=='Darwin'",
        "scipy",
        "matplotlib",
        "netCDF4",
        "h5netcdf",
        "datetime",
        "utm",
        "rasterio",
        "scikit-image",
        "pandas",
        "geopandas",
        "shapely",
        "cdsapi",
        "prompt_toolkit",
        "siphon",
        "cfgrib",
        "xarray",
        "rioxarray",
        "requests",
        "contextily",
        "osmnx",
        "geojsoncontour",
        "folium",
    ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
)



from collections import OrderedDict

import numpy as np
from numba import float64
from numba.experimental import jitclass

phys_spec = OrderedDict()
phys_spec["Kappa_h"] = float64  # Horizontal diffusion coefficient
phys_spec["Kappa_v"] = float64  # Vertical diffusion coefficient
phys_spec["g"] = float64  # Gravitational acceleration
phys_spec["mu"] = float64  # Viscosity of air


@jitclass(phys_spec)
class PhysicalParameters:
    def __init__(self, Kappa_h=100, Kappa_v=10, g=9.81, mu=18.5e-6):

        if Kappa_h < 0:
            raise ValueError("In PhysicalParameters, Kappa_h must be positive")
        self.Kappa_h = np.float64(Kappa_h)

        if Kappa_v < 0:
            raise ValueError("In PhysicalParameters, Kappa_v must be positive")
        self.Kappa_v = np.float64(Kappa_v)

        if g <= 0:
            raise ValueError("In PhysicalParameters, g must be positive")
        self.g = np.float64(g)

        if mu <= 0:
            raise ValueError("In PhysicalParameters, mu must be positive")
        self.mu = np.float64(mu)

    def describe(self):
        print("Physical parameters for AshDisperse")
        print("  Horizontal diffusion coefficient Kappa_h = ", self.Kappa_h, " m^2/s")
        print("  Vertical diffusion coefficient Kappa_v = ", self.Kappa_v, " m^2/s")
        print("  Gravitational acceleration g = ", self.g, " m/s^2")
        print("  Viscosity of air mu = ", self.mu, " kg/m/s")
        print("********************")


# pylint: disable=E1101
PhysicalParameters_type = PhysicalParameters.class_type.instance_type



import numpy as np
from numba import complex128, float64, int64, jit

# from numba.pycc import CC
from numba.types import Tuple

# cc = CC('cheb')
# cc.verbose = True


# @cc.export('cheb', Tuple((float64[::1], float64[:, ::1], float64[:, ::1], float64[:, ::1]))(int64))
@jit(
    Tuple((float64[::1], float64[:, ::1], float64[:, ::1], float64[:, ::1]))(int64),
    nopython=True,
    cache=True,
)
def cheb(N):
    t = np.zeros((N), dtype=np.float64)
    x = np.zeros((N), dtype=np.float64)
    T = np.zeros((N, N), dtype=np.float64)
    dT = np.zeros((N, N), dtype=np.float64)
    d2T = np.zeros((N, N), dtype=np.float64)

    t[:] = np.pi * np.arange(0, N) / (N - 1)
    t = t[::-1]
    sint = np.sin(t)
    cost = np.cos(t)
    sin2t = sint * sint
    sin3t = sin2t * sint
    x = cost
    a = np.ones((N), dtype=np.float64)
    a[1::2] = -1.0
    T[:, 0] = 1.0
    dT[:, 0] = 0.0
    d2T[:, 0] = 0.0
    T[:, 1] = x
    dT[:, 1] = 1.0
    d2T[:, 1] = 0.0
    T[:, 2] = 2.0 * x * x - 1.0
    dT[:, 2] = 4.0 * x
    d2T[:, 2] = 4.0
    for n in range(3, N):
        T[1:-1, n] = np.cos(n * t[1:-1])
        T[0, n] = a[n]
        T[-1, n] = 1.0
        dT[1:-1, n] = n * np.sin(n * t[1:-1]) / sint[1:-1]
        dT[0, n] = -a[n] * n * n
        dT[-1, n] = n * n
        d2T[1:-1, n] = (
            n * cost[1:-1] * np.sin(n * t[1:-1]) / sin3t[1:-1]
            - n * n * np.cos(n * t[1:-1]) / sin2t[1:-1]
        )
        d2T[0, n] = a[n] * (n**4 - n**2) / 3
        d2T[-1, n] = (n**4 - n**2) / 3
    return x, T, dT, d2T


# @cc.export('ChebMat', complex128[:, ::1](int64, float64[::1]))
@jit(complex128[:, ::1](int64, float64[::1]), nopython=True, cache=True)
def ChebMat(N, x):
    """Evaluation at +1 of a function defined by Chebyshev coefficients.

    Args:
        coeffs: the Chebyshev coefficients as an array of complex numbers.

    Returns:
        The function evaluated at +1.
    """
    T = np.zeros((x.size, N), dtype=complex128)
    acos = np.arccos(x)
    for k in range(0, N):
        T[:, k] = np.cos(k * acos)
    return T


# @cc.export('cheb_val_p1', complex128(complex128[::1]))
@jit(complex128(complex128[::1]), nopython=True, cache=True)
def cheb_val_p1(coeffs):
    """Evaluates at +1 of a function defined by Chebyshev coefficients.

    Args:
        coeffs: the Chebyshev coefficients as an array of complex numbers.

    Returns:
        The function evaluated at +1.
    """
    return np.sum(coeffs)


# @cc.export('cheb_dif_p1', complex128(complex128[::1]))
@jit(complex128(complex128[::1]), nopython=True, cache=True)
def cheb_dif_p1(coeffs):
    """Evaluates at +1 of the first derivative of a function
        defined by Chebyshev coefficients.

    Args:
        coeffs: the Chebyshev coefficients as an array of complex numbers.

    Returns:
        The derivative of the function evaluated at +1.
    """
    n2 = np.arange(len(coeffs), dtype=np.complex128) ** 2
    df = np.dot(n2, np.ascontiguousarray(coeffs))
    return df


# @cc.export('cheb_val_m1', complex128(complex128[::1]))
@jit(complex128(complex128[::1]), nopython=True, cache=True)
def cheb_val_m1(coeffs):
    """Evaluates at -1 of a function defined by Chebyshev coefficients.

    Args:
        coeffs: the Chebyshev coefficients as an array of complex numbers.

    Returns:
        The function evaluated at -1.
    """
    a = np.ones((len(coeffs)))
    a[1::2] = -1.0
    f = np.dot(a.astype(np.complex128), np.ascontiguousarray(coeffs))
    return f


# @cc.export('cheb_dif_m1', complex128(complex128[::1]))
@jit(complex128(complex128[::1]), nopython=True, cache=True)
def cheb_dif_m1(coeffs):
    """Evaluation at -1 of the first derivative of a function
        defined by Chebyshev coefficients.

    Args:
        coeffs: the Chebyshev coefficients as an array of complex numbers.

    Returns:
        The derivative of the function evaluated at -1.
    """
    N = len(coeffs)
    n2 = np.arange(N) ** 2
    a = np.ones((N), dtype=np.int64)
    a[0::2] = -1.0
    b = a * n2
    func_deriv = np.dot(b.astype(np.complex128), np.ascontiguousarray(coeffs))
    return func_deriv


# @cc.export('convergedCoeffs', complex128[::1](complex128[::1], float64))
@jit(complex128[::1](complex128[::1], float64), nopython=True, cache=True)
def convergedCoeffs(coeffs, eps):

    N = coeffs.size

    a = np.abs(coeffs)
    max_a = np.amax(a)

    if max_a > 0:
        neg = (a < eps) * 1

        if neg[-1] and neg[-2]:
            n = np.arange(0, N)
            nGood = n[neg == 0]
            Ntrunc = nGood[-1] + 1
        else:
            Ntrunc = N

        new_coeffs = coeffs[:Ntrunc].flatten()

    else:
        new_coeffs = np.zeros((1), dtype=np.complex128)

    return new_coeffs


# if __name__ == "__main__":
#     cc.compile()



# -*- coding: utf-8 -*-
import datetime
import os
import tkinter as tk
from tkinter.filedialog import askopenfilename

import numpy as np
from siphon.catalog import TDSCatalog

from ..config import config
from ..met import load_met, met, save_met
from ..params import (
    GrainParameters,
    MetParameters,
    ModelParameters,
    OutputParameters,
    Parameters,
    PhysicalParameters,
    SolverParameters,
    SourceParameters,
    load_parameters,
    save_parameters,
)
from ..queryreport import (
    print_text,
    print_title,
    query_change_value,
    query_choices,
    query_datetime,
    query_latlon,
    query_met_file,
    query_set_value,
    query_yes_no,
)
from ..utilities import latlon_point_to_utm_code, string_to_datetime


def print_grain_parameters(grain_params):
    """Prints the grain classes in GrainParameters object.

    This function prints a nicely formatted description of a GrainParameters
    object.

    Args:
        grain_params: a GrainParameters instance.

    Raises:
        ValueError: if grain_params in not a GrainParameters instance.
    """
    if not isinstance(grain_params, GrainParameters):
        raise ValueError(
            "in print_grain_parameters," + " argument must be a GrainParameters object"
        )

    print_title("Grain parameters for AshDisperse")
    print_text("  Number of grain classes, N_grains = {}".format(grain_params.bins))
    for j, (diam, density, prop) in enumerate(
        zip(grain_params.diameter, grain_params.density, grain_params.proportion)
    ):
        print_text("  Grain class {}".format(j))
        print_text("    Grain diameter = {} m".format(diam))
        print_text("    Grain density = {} kg/m^3".format(density))
        print_text("    Proportion = {}".format(prop))
    print_text("********************")


def print_model_parameters(model_params):
    """Prints the model parameters in the ModelParameters object.

    This function prints a nicely formatted description of a ModelsParameters
    object.

    Args:
        model_params: a ModelParameters instance.

    Raises:
        ValueError: if grain_params in not a GrainParameters instance.
    """

    print_title("Model parameters for AshDisperse")
    print_text("  Settling speed scale = {} ".format(model_params.SettlingScale))
    print_text("  Velocity ratio = {} ".format(model_params.Velocity_ratio))
    print_text("  x and y scale = {} ".format(model_params.xyScale))
    print_text("  Lx = {} ".format(model_params.Lx))
    print_text("  Ly = {} ".format(model_params.Ly))
    print_text("  concentration scale = {} ".format(model_params.cScale))
    print_text("  source flux scale = {} ".format(model_params.QScale))
    print_text("  Peclet number = {} ".format(model_params.Peclet_number))
    print_text("  Diffusion ratio = {} ".format(model_params.Diffusion_ratio))
    print_text("********************")


def print_output_parameters(output_params):
    print_title("Output parameters for AshDisperse")
    print_text("  Altitudes = {} m ".format(output_params.altitudes))
    print_text("********************")


def print_physical_parameters(physical_params):
    print_title("Physical parameters for AshDisperse")
    print_text(
        "  Horizontal diffusion coefficient Kappa_h = {} m^2/s ".format(
            physical_params.Kappa_h
        )
    )
    print_text(
        "  Vertical diffusion coefficient Kappa_v = {} m^2/s ".format(
            physical_params.Kappa_v
        )
    )
    print_text("  Gravitational acceleration g = {} m/s^2 ".format(physical_params.g))
    print_text("  Viscosity of air mu = {} kg/m/s ".format(physical_params.mu))
    print_text("********************")


def print_solver_parameters(solver_params):
    print_title("Solver parameters for AshDisperse")
    print_text("  Dimensionless domain size in x, domX = {}".format(solver_params.domX))
    print_text("  Dimensionless domain size in y, domY = {}".format(solver_params.domY))
    print_text(
        "  Minimum resolution in z, minN = {}, (minN_log2 = {})".format(
            solver_params.minN, solver_params.minN_log2
        )
    )
    print_text(
        "  Maximum resolution in z, maxN = {}, (maxN_log2 = {})".format(
            solver_params.maxN, solver_params.maxN_log2
        )
    )
    print_text("  Number of Chebyshev iterates = {}".format(solver_params.chebIts))
    print_text(
        "  Tolerance for Chebyshev series, epsilon = {}".format(solver_params.epsilon)
    )
    print_text(
        "  Resolution in x, Nx = {}, (Nx_log2 = {})".format(
            solver_params.Nx, solver_params.Nx_log2
        )
    )
    print_text(
        "  Resolution in y, Ny = {}, (Ny_log2 = {})".format(
            solver_params.Ny, solver_params.Ny_log2
        )
    )
    print_text("********************")


def print_source_parameters(source_params):
    """Print source parameters.

    Args:
        source_params (SourceParameters): A SourceParameters object
    """
    if not isinstance(source_params, SourceParameters):
        raise ValueError(
            "input to print_source_parameters must be a 'SourceParameters'"
            + " instance"
        )
    print_title("Source parameters for AshDisperse")
    print_text("  Mass eruption rate MER = {} kg/s".format(source_params.MER))
    print_text(
        "  Eruption duration = {} ".format(
            str(datetime.timedelta(seconds=source_params.MER))
        )
    )
    print_text("  Plume height H = {} m".format(source_params.PlumeHeight))
    print_text("  Gaussian source radius = {} m".format(source_params.radius))
    print_text(
        "  Suzuki emission profile k-parameter = {}".format(source_params.Suzuki_k)
    )
    print_text(
        "  Suzuki emission profile peak-parameter = {}".format(
            source_params.Suzuki_peak
        )
    )
    print_text("********************")


def set_grain_parameters(params_set, printout=True):
    """Set grain parameters for AshDisperse.

    The Parameters class object stores the parameters required for the model.
    set_grain_parameters runs a series of command line inputs to initialize
    the GrainParameters element of the Parameters object.

    Args:
        params_set (Parameters): The Parameters object to initialize.
        printout (bool, optional): Print GrainParameters after setup.

    Returns:
        params_set (Parameters): An updated Parameters object.

    Raises:
        ValueError: If 'params_set' is not a Parameters object.
                    If user inputs a negative diameter.
                    If user inputs a negative density.
                    If user inputs a proportion outside of [0,1).
                    If user inputs a proportion that is too large.
                    If grain parameters are not valid according to validate()
                        method of GrainParameters.
    """
    if not isinstance(params_set, Parameters):
        raise ValueError(
            "input to set_grain_parameters must be a 'Parameters' instance"
        )

    params_set.grains = GrainParameters()
    combined_prop = 0
    while combined_prop < 1:
        print_title("Add grain class")
        grain_diam = query_set_value("  Grain diameter (m): ", answer_type=float)
        if grain_diam <= 0:
            raise ValueError("Grain diameter must be positive")
        grain_density = query_set_value("  Grain density (kg/m^3): ", answer_type=float)
        if grain_density <= 0:
            raise ValueError("Grain density must be positive")
        grain_proportion = query_set_value(
            "  Grain class proportion (0--{}): ".format(1 - combined_prop),
            answer_type=float,
        )
        if grain_proportion <= 0 or grain_proportion > 1:
            raise ValueError("Grain proportion must be between 0 and 1")
        combined_prop += grain_proportion
        params_set.grains.add_grain(grain_diam, grain_density, grain_proportion)
    valid_grain = params_set.grains.validate()
    if not valid_grain:
        raise ValueError("Grain classes not valid")
    if printout:
        print_grain_parameters(params_set.grains)
    return params_set


def set_output_parameters(params_set, printout=False):
    """Set output parameters for AshDisperse.

    The Parameters class object stores the parameters required for the model.
    set_output_parameters runs a series of command line inputs to initialize
    the OutputParameters element of the Parameters object.

    Args:
        params_set (Parameters): The Parameters object to initialize.
        printout (bool, optional): Print GrainParameters after setup.

    Returns:
        params_set (Parameters): An updated Parameters object.

    Raises:
        ValueError: If 'params_set' is not a Parameters object

    TODO (Mark): Add validation of parameters.
    """
    if not isinstance(params_set, Parameters):
        raise ValueError(
            "input to set_output_parameters must be a 'Parameters' instance"
        )
    params_set.output.start = query_change_value(
        "  Lower altitude", default=params_set.output.start, answer_type=float
    )
    params_set.output.stop = query_change_value(
        "  Upper altitude", default=params_set.output.stop, answer_type=float
    )
    params_set.output.step = query_change_value(
        "  Altitude step", default=params_set.output.step, answer_type=float
    )

    params_set.output.set_altitudes()
    params_set.output.ChebMats(params_set.solver.maxN, params_set.source.PlumeHeight)

    if printout:
        print_output_parameters(params_set.output)

    return params_set


def set_parameters():
    """Interface to set Parameters for AshDisperse.

    The Parameters class object stores the parameters required for the model.
    set_parameters runs a series of command line inputs to initialize
    the Parameters object.

    Args:
        None

    Returns:
        params_set (Parameters): A Parameters object.

    """
    print_title("Set parameters for AshDisperse")
    params_set = Parameters()

    print_title("Set source parameters")
    params_set = set_source_parameters(params_set)

    print_title("Set grain parameters")
    params_set = set_grain_parameters(params_set)

    print_title("Set solver parameters")
    params_set.solver = SolverParameters()
    print_solver_parameters(params_set.solver)
    solver_change = query_yes_no("Change solver parameters?", default="no")
    if solver_change:
        params_set = set_solver_parameters(params_set)

    print_title("Set physical parameters")
    params_set.physical = PhysicalParameters()
    print_physical_parameters(params_set.physical)
    physical_change = query_yes_no("Change physical parameters?", default="no")
    if physical_change:
        params_set = set_physical_parameters(params_set)

    print_title("Set output parameters")
    params_set.output = OutputParameters(
        start=0,
        stop=2 * params_set.source.PlumeHeight,
        step=2 * params_set.source.PlumeHeight / 10,
    )
    print_output_parameters(params_set.output)
    output_change = query_yes_no("Change output parameters?", default="no")
    if output_change:
        params_set = set_output_parameters(params_set)
    else:
        params_set.output.set_altitudes()
        params_set.output.ChebMats(
            params_set.solver.maxN, params_set.source.PlumeHeight
        )

    return params_set


def set_physical_parameters(params_set, printout=False):
    """Set physical parameters for AshDisperse.

    The Parameters class object stores the parameters required for the model.
    set_physical_parameters runs a series of command line inputs to initialize
    the PhysicalParameters element of the Parameters object.

    Args:
        params_set (Parameters): The Parameters object to initialize.
        printout (bool, optional): Print GrainParameters after setup.

    Returns:
        params_set (Parameters): An updated Parameters object.

    Raises:
        ValueError: If 'params_set' is not a Parameters object

    TODO (Mark): Add validation of parameters.
    """
    if not isinstance(params_set, Parameters):
        raise ValueError(
            "input to set_physical_parameters must be a 'Parameters' instance"
        )

    params_set.physical.Kappa_h = query_change_value(
        "  Horizontal diffusion coefficient, kappa_h",
        default=params_set.physical.Kappa_h,
        answer_type=float,
    )
    params_set.physical.Kappa_v = query_change_value(
        "  Vertical diffusion coefficient, kappa_v",
        default=params_set.physical.Kappa_v,
        answer_type=float,
    )
    params_set.physical.g = query_change_value(
        "  Gravitational acceleration, g",
        default=params_set.physical.g,
        answer_type=float,
    )
    params_set.physical.mu = query_change_value(
        "  Viscosity of air, mu", default=params_set.physical.mu, answer_type=float
    )

    if printout:
        print_solver_parameters(params_set.solver)

    return params_set


def set_source_parameters(params_set):
    """Set source parameters for AshDisperse.

    The Parameters class object stores the parameters required for the model.
    set_source_parameters runs a series of command line inputs to initialize
    the SourceParameters element of the Parameters object.

    Args:
        params_set (Parameters): The Parameters object to initialize.

    Returns:
        params_set (Parameters): An updated Parameters object.

    Raises:
        ValueError: If 'params_set' is not a Parameters object
    """
    if not isinstance(params_set, Parameters):
        raise ValueError(
            "input to set_source_parameters must be a 'Parameters' instance"
        )

    name, lat, lon = query_latlon()

    params_set.source = SourceParameters(
        lat, lon, latlon_point_to_utm_code(lat, lon), name=name
    )

    params_set.source.MER = query_change_value(
        "  Mass eruption rate", default=params_set.source.MER, answer_type=float
    )
    params_set.source.duration = query_change_value(
        "  Eruption duration", default=params_set.source.duration, answer_type=float
    )
    params_set.source.PlumeHeight = query_change_value(
        "  Plume height", default=params_set.source.PlumeHeight, answer_type=float
    )
    params_set.source.radius = query_change_value(
        "  Gaussian source radius", default=params_set.source.radius, answer_type=float
    )
    suzuki_choice = query_choices(
        "  Select Suzuki emission profile parameter: ",
        choices=["k", "peak"],
        default="k",
    )
    if suzuki_choice == "k":
        params_set.source.Suzuki_k = query_change_value(
            "  Suzuki emission profile k-parameter",
            default=params_set.source.Suzuki_k,
            answer_type=float,
        )
        params_set.source.peak_from_k()
    else:
        params_set.source.Suzuki_peak = query_change_value(
            "  Suzuki emission profile peak-parameter",
            default=params_set.source.Suzuki_peak,
            answer_type=float,
        )
        params_set.source.k_from_peak()
    return params_set


def set_solver_parameters(params_set, printout=False):
    """Set solver parameters for AshDisperse.

    The Parameters class object stores the parameters required for the model.
    set_solver_parameters runs a series of command line inputs to initialize
    the SolverParameters element of the Parameters object.

    Args:
        params_set (Parameters): The Parameters object to initialize.
        printout (bool, optional): Print GrainParameters after setup.

    Returns:
        params_set (Parameters): An updated Parameters object.

    Raises:
        ValueError: If 'params_set' is not a Parameters object

    TODO (Mark): Add validation of parameters.
    """
    if not isinstance(params_set, Parameters):
        raise ValueError(
            "input to set_solver_parameters must be a 'Parameters' instance"
        )

    params_set.solver.domX = query_change_value(
        "  Dimensionless domain size in x, domX",
        default=params_set.solver.domX,
        answer_type=float,
    )
    params_set.solver.domY = query_change_value(
        "  Dimensionless domain size in y, domY",
        default=params_set.solver.domY,
        answer_type=float,
    )
    params_set.solver.minN_log2 = query_change_value(
        "  Minimum resolution in z, log2(minN)",
        default=params_set.solver.minN_log2,
        answer_type=int,
    )
    params_set.solver.maxN_log2 = query_change_value(
        "  Maximum resolution in z, log2(maxN)",
        default=params_set.solver.maxN_log2,
        answer_type=int,
    )
    params_set.solver.epsilon = query_change_value(
        "  Tolerance for Chebyshev series, epsilon",
        default=params_set.solver.epsilon,
        answer_type=float,
    )
    params_set.solver.Nx_log2 = query_change_value(
        "  Resolution in x, log2(Nx)",
        default=params_set.solver.Nx_log2,
        answer_type=int,
    )
    params_set.solver.Ny_log2 = query_change_value(
        "  Resolution in y, log2(Ny)",
        default=params_set.solver.Ny_log2,
        answer_type=int,
    )
    params_set.solver.chebIts = (
        params_set.solver.maxN_log2 - params_set.solver.minN_log2 + 1
    )
    params_set.solver.minN = 2**params_set.solver.minN_log2
    params_set.solver.maxN = 2**params_set.solver.maxN_log2
    params_set.solver.Nx = 2**params_set.solver.Nx_log2
    params_set.solver.Ny = 2**params_set.solver.Ny_log2

    params_set.solver.rk = query_yes_no("  Use Runge-Kutta solver?", default="no")
    if params_set.solver.rk:
        params_set.solver.rtol = query_change_value(
            "  Runge-Kutta relative error tolerance, rtol",
            default=params_set.solver.rtol,
            answer_type=float,
        )
        params_set.solver.maxStep = query_change_value(
            "  Runge-Kutta maximum step size, maxStep",
            default=params_set.solver.maxStep,
            answer_type=float,
        )

    if printout:
        print_solver_parameters(params_set.solver)

    return params_set


def set_met(params_set, source="interface", **kwargs):

    if source == "interface":
        if "gui" in kwargs.keys():
            gui = kwargs["gui"]
        else:
            gui = False
        met_set = _set_met_interface(params_set, gui=gui)
    elif source == "local":
        if "met_filename" not in kwargs.keys():
            raise RuntimeError(
                "In met_set, 'met_filename' must be given if source='local'"
            )
        met_filename = kwargs["met_filename"]
        met_set = _get_local_netcdf_met(
            met_filename, params_set.source.latitude, params_set.source.longitude
        )
    elif source == "gfs":
        if "datetime" not in kwargs.keys():
            raise RuntimeError("In met_set, 'date' must be given if source='gfs'")
        date = kwargs["datetime"]
        met_datetime = string_to_datetime(date)
        met_set = _get_gfs_met(
            params_set.source.latitude, params_set.source.longitude, met_datetime
        )
    elif source == "ecmwf":
        if "datetime" not in kwargs.keys():
            raise RuntimeError("In met_set, 'date' must be given if source='ecmwf'")
        date = kwargs["datetime"]
        met_datetime = string_to_datetime(date)
        if "met_filename" in kwargs.items():
            met_filename = kwargs["met_filename"]
        else:
            met_filename = "./MetData/era5_download.nc"
        met_set = _get_ecmwf_met(
            params_set.source.latitude,
            params_set.source.longitude,
            met_datetime,
            met_filename=met_filename,
        )

    params_set = _set_met_parameters(params_set, met_set)

    return params_set, met_set


def _set_met_parameters(params_set, met_set):

    wind_speed = met_set.max_wind_speed(params_set.source.PlumeHeight)

    settling_speed = met_set.calculate_settling_speed_value(params_set, 0.0)

    params_set.met = MetParameters(wind_speed, settling_speed)

    params_set.model = ModelParameters()
    params_set.model.from_params(
        params_set.solver,
        params_set.met,
        params_set.source,
        params_set.grains,
        params_set.physical,
    )

    return params_set


def _set_met_interface(params_set, gui=False):
    """Interface to set MetData for AshDisperse.

    The MetData class object stores the meteorological data required for the
    model.
    set_met runs a series of user inputs to initialize the MetData object.
    Additionally, setting meteorological data finalizes the Parameters object
    that stores parameters for the model.

    Args:
        params_set (Parameters): A Parameters object containing parameters for
                                 the model.
        gui (bool, optional): Use a tkinter gui; defaults to False.

    Returns:
        met_set (MetData): A MetData object containing meteorological data.
        params_set (Parameters): An updated Parameters object.

    Raises:
        OSError: If user inputted file not found.
    """
    print_title("Set up meteorological inputs")
    met_source = int(
        query_choices(
            "Source of meteorological data: \n"
            + "   0 -> local netcdf file \n"
            + "   1 -> download ERA5 data \n"
            + "   2 -> standard atmosphere \n"
            + "   3 -> GFS forecast \n"
            + "   4 -> GFS archive \n",
            choices=["0", "1", "2", "3", "4"],
            default="0",
        )
    )

    if met_source == 0:

        if gui:
            tk.Tk().withdraw()
            met_file = askopenfilename(
                initialdir="./",
                title="Select met file (netCDF)",
                filetypes=(("netCDF", "*.nc"), ("all files", "*.*")),
            )
        else:
            met_file = query_met_file()

        met_set = _get_local_netcdf_met(
            met_file, params_set.source.latitude, params_set.source.longitude
        )

    elif met_source == 1:
        met_file = query_change_value(
            "File name for download",
            default="./MetData/era5_download.nc",
            answer_type=str,
        )
        met_datetime = query_datetime(
            "Date and time for meteorological data (as yyyy-mm-dd hh:mm)"
        )

        met_set = _get_ecmwf_met(
            params_set.source.latitude,
            params_set.source.longitude,
            met_datetime,
            met_filename=met_file,
        )
    elif met_source == 2:
        print("Using Standard Atmosphere")
        met_set = met.MetData()
    elif met_source == 3:
        met_datetime = query_datetime(
            "Date and time for meteorological data (as yyyy-mm-dd hh:mm)"
        )
        met_set = _get_gfs_met(
            params_set.source.latitude, params_set.source.longitude, met_datetime
        )
    elif met_source == 4:
        met_datetime = query_datetime(
            "Date and time for GFS forecast cycle (as yyyy-mm-dd hh:00)"
        )
        forecast_hr = query_set_value(
            "Forecast hour (hours ahead of forecast cycle)", answer_type=int
        )
        met_set = _get_gfs_archive(
            params_set.source.latitude,
            params_set.source.longitude,
            met_datetime,
            forecast_hr,
        )

    return met_set


def _get_local_netcdf_met(met_file, latitude, longitude):
    if not os.path.isfile(met_file):
        raise OSError("Met file {} does not exist".format(met_file))
    cdf_data = met.netCDF(met_file, lat=latitude, lon=longitude)
    cdf_data.extract()
    met_set = met.MetData()
    met_set = met.netCDF_to_Met(met_set, cdf_data)
    return met_set


def _get_ecmwf_met(
    latitude, longitude, met_datetime, met_filename="./MetData/era5_download.nc"
):
    cdf_data = met.netCDF(met_filename, lat=latitude, lon=longitude)
    cdf_data.download(datetime=met_datetime)
    cdf_data.extract()
    met_set = met.MetData()
    met_set = met.netCDF_to_Met(met_set, cdf_data)
    return met_set


def _get_gfs_met(latitude, longitude, met_datetime):
    gfs = TDSCatalog(
        "http://thredds.ucar.edu/thredds/catalog/grib/NCEP/GFS/Global_0p25deg/catalog.xml"
        "?dataset=grib/NCEP/GFS/Global_0p25deg/Best"
    )
    gfs_ds = gfs.datasets[0]
    ncss = gfs_ds.subset()
    query = ncss.query()
    query.lonlat_point(longitude, latitude)
    query.time(met_datetime)
    query.variables(
        "Temperature_isobaric",
        "Geopotential_height_isobaric",
        "u-component_of_wind_isobaric",
        "v-component_of_wind_isobaric",
    ).accept("netcdf")
    gfs_data = ncss.get_data(query)
    met_set = met.MetData()
    met_set = met.gfs_to_Met(met_set, gfs_data)
    return met_set


def _get_gfs_archive(latitude, longitude, cycle_datetime, forecast_hr):
    gfs_data = met.gfs_archive(
        cycle_datetime, forecast_hr, lat=latitude, lon=longitude
    ).profiles()

    met_set = met.MetData()
    met_set = met.gfs_archive_to_Met(met_set, gfs_data)
    return met_set


def save_inputs(params, met, filename="AshDisperse"):
    save_parameters(params, file=f"{filename}.params.txt")
    save_met(met, file=f"{filename}.met.npz")


def load_inputs(filename):
    params = load_parameters(f"{filename}.params.txt")
    met = load_met(f"{filename}.met.npz")

    params = _set_met_parameters(params, met)

    return params, met



import numpy as np
from scipy.fft import fft2


def source_xy_dimless_flat(x, y, p):
    xx, yy = np.meshgrid(x, y)
    pi = np.pi

    f = np.zeros((len(y), len(x)), dtype=np.float64)
    fxy = np.zeros((len(y), len(x), p.grains.bins), dtype=np.float64)
    fxy_f = np.zeros((len(y), len(x), p.grains.bins), dtype=np.complex128)

    dx = (x[1] - x[0])
    dy = (y[1] - y[0])

    Lr = p.solver.domX/p.solver.domY

    rr2 = Lr*np.power(xx, 2) + np.power(yy, 2)/Lr

    for j in range(p.grains.bins):
        f[:, :] = 1/np.pi
        sigma2 = (p.source.radius**2/p.model.xyScale[j]**2
                  * pi**2/(p.solver.Lx*p.solver.Ly))
        f[rr2 > sigma2] = 0.0

        fxy[:, :, j] = f
        fxy_f[:, :, j] = fft2(fxy[:, :, j])*dx*dy

    return fxy, fxy_f


def source_xy_dimless(x, y, p):
    xx, yy = np.meshgrid(x, y)
    pi = np.pi

    fxy = np.zeros((len(y), len(x), p.grains.bins), dtype=np.float64)
    fxy_f = np.zeros((len(y), len(x), p.grains.bins), dtype=np.complex128)

    dx = (x[1] - x[0])
    dy = (y[1] - y[0])

    Lr = p.solver.domX/p.solver.domY

    rr2 = Lr*np.power(xx, 2) + np.power(yy, 2)/Lr

    for j in range(p.grains.bins):

        sigma = (p.source.radius/3./p.model.xyScale[j]
                 * pi/np.sqrt(p.solver.domX*p.solver.domY))

        var = np.power(sigma, 2)

        fxy[:, :, j] = np.exp(-0.5*rr2/var)/(2.0*pi)
        fxy_f[:, :, j] = fft2(fxy[:, :, j])*dx*dy
    return fxy, fxy_f



from .epsg_defs import webmerc, wgs84

import numpy as np
import rasterio as rio
import contextily as ctx
import geopandas as gpd
import pyproj as Proj
import osmnx as osm

from shapely import wkt
from shapely.geometry import Point, Polygon, MultiPoint, MultiPolygon, LineString, MultiLineString, LinearRing, mapping
from shapely.ops import transform, linemerge, unary_union, polygonize
from shapely.affinity import translate

from functools import partial

import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.collections import PatchCollection
import matplotlib.colors as colors

from requests.exceptions import HTTPError

from ..utilities import nice_round_down

def latlon_to_utm_epsg(lat, lon):
    offset = int(round((183+lon)/6.0))
    return 32600+offset if (lat > 0) else 32700+offset

def expand_axes(ax,abs_x=0,abs_y=0):
    x0, x1, y0, y1 = ax.axis()
    x0 = x0-abs_x/2.
    x1 = x1+abs_x/2.
    y0 = y0-abs_y/2.
    y1 = y1+abs_y/2.
    ax.set_xlim(x0,x1)
    ax.set_ylim(y0,y1)
    return

def set_min_axes(ax,min_width=50000,min_height=50000):
    x0, x1, y0, y1 = ax.axis()
    x0=x0
    x1=x1
    y0=y0
    y1=y1
    w = (x1-x0)
    h = (y1-y0)
    ew = max(0,min_width-w)
    eh = max(0,min_height-h)
    ax.set_xlim(x0-ew/2,x1+ew/2)
    ax.set_ylim(y0-eh/2,y1+eh/2)
    return

def ax_ticks(ax, x0, y0):
    xmin, xmax, ymin, ymax = ax.axis()
    width = xmax - xmin
    height = ymax - ymin
    
    xticklabels = np.linspace(-nice_round_down(width)/2, nice_round_down(width)/2, 5)
    yticklabels = np.linspace(-nice_round_down(height)/2, nice_round_down(height)/2, 5)

    xticks = x0 + xticklabels
    yticks = y0 + yticklabels

    ax.set_xticks(xticks)
    ax.set_xticklabels(xticklabels/1e3)

    ax.set_yticks(yticks)
    ax.set_yticklabels(yticklabels/1e3)

    ax.set_xlabel('Easting (km from vent)')
    ax.set_ylabel('Northing (km from vent)')

    return ax

def set_figure_size(fig, ax):
    xmin, xmax, ymin, ymax = ax.axis()
    ax_width = xmax - xmin
    ax_height = ymax - ymin

    fig_height = fig.get_figheight()
    fig_width = fig.get_figwidth()
    fig.set_figwidth(fig_height*ax_width/ax_height)

    return fig

def add_basemap(ax,zoom,provider='Stamen',style='TerrainBackground',zorder=0):
    valid_providers = ctx.providers.keys()
    if provider not in valid_providers:
        raise ValueError("in add_basemap: provider must be one of %r." % valid_providers)
    valid_styles = ctx.providers[provider].keys()
    if style not in valid_styles:
        raise ValueError("in add_basemap: style must be one of %r." % valid_styles)
    ctx.add_basemap(ax,zoom=zoom,source=ctx.providers[provider][style],zorder=zorder)
    return ax

def add_stamen_basemap(ax, max_zoom=13, max_tiles=256, stamen='terrain', zorder=0): 
    validStamen = {'terrain','background','labels','lines'}
    if stamen not in validStamen:
        raise ValueError("in add_stamen_basemap: stamen must be one of %r." % validStamen)
    if stamen=='background':
        url='http://tile.stamen.com/terrain-background/{z}/{x}/{y}.png'
    elif stamen=='labels':
        url='http://tile.stamen.com/terrain-labels/{z}/{x}/{y}.png'
    elif stamen=='lines':
        url='http://tile.stamen.com/terrain-lines/{z}/{x}/{y}.png'
    else:
        url='http://tile.stamen.com/terrain/{z}/{x}/{y}.png'
    xmin, xmax, ymin, ymax = ax.axis()
    zoom = max_zoom
    while zoom>0:
        ntiles = ctx.howmany(xmin, ymin, xmax, ymax, zoom, verbose=False)
        if ntiles<=max_tiles:
            try:
                basemap, extent = ctx.bounds2img(xmin, ymin, xmax, ymax, zoom=zoom, source=url)
                break
            except HTTPError:
                zoom -= 1
                continue
        else:
            zoom -= 1
    if zoom>0:
        ax.imshow(basemap, extent=extent, interpolation='bilinear',zorder=zorder)
        # restore original x/y limits 
        ax.axis((xmin, xmax, ymin, ymax))
    return ax

def add_esri_basemap(ax, variant='WorldPhysical', zorder=0):
    esri = ctx.providers.Esri
    variants = list(esri.keys())
    if variant not in variants:
        raise ValueError("in add_esri_basemap: variant must be one of {}.".format(variants))
    url = esri[variant].url.replace('{variant}',esri[variant].variant)
    try:
        zoom = esri[variant].max_zoom
    except:
        zoom = 20
    while zoom>0:
        try:
            basemap, extent = ctx.bounds2img(xmin, ymin, xmax, ymax, zoom=zoom, source=url)
            break
        except HTTPError:
            zoom -= 1
            continue
    if zoom>0:
        ax.imshow(basemap, extent=extent, interpolation='bilinear',zorder=zorder)
        # restore original x/y limits 
        ax.axis((xmin, xmax, ymin, ymax))
    return ax

def stamen_zoom_level(fig, ax, max_level=13, dpi=300):
    x0, x1, y0, y1 = ax.axis()

    bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    width = bbox.width

    t = Proj.Transformer.from_crs(webmerc['init'], wgs84['init'])

    rightmid = t.transform(x1, 0.5*(y0+y1))
    centermid = t.transform(0.5*(x0+x1), 0.5*(y0+y1))
    leftmid = t.transform(x0,0.5*(y0+y1))

    anglewidth1 = rightmid[0] - centermid[0]
    if anglewidth1<0:
        anglewidth1 = anglewidth1+360.
    
    anglewidth2 = rightmid[1] - centermid[1]
    if anglewidth2<0:
        anglewidth2 = anglewidth2+360.
    
    anglewidth = anglewidth1+anglewidth2

    zoom = np.log2((360./anglewidth)*(width*dpi/256.))
    zoom = int(np.floor(zoom))
    if zoom > max_level:
        zoom = max_level
    return zoom


def _geodesic_point(poiGDF,xy_displ):
    lat = poiGDF.y
    lon = poiGDF.x
    pointsGDF = gpd.GeoDataFrame()
    pointsGDF['geometry']=None
    pointsGDF['distance']=None
    pointsGDF.crs = poiGDF.crs
    aeqd_proj = '+proj=aeqd +lat_0={lat} +lon_0={lon} +x_0=0 +y_0=0'
    project = partial(
        Proj.transform,
        Proj.Proj(aeqd_proj.format(lat=lat,lon=lon)),
        Proj.Proj(poiGDF.crs)
        )
    for j,[x,y] in enumerate(xy_displ):
        pt = translate(Point(0,0),xoff=x,yoff=y)
        pointsGDF.loc[j,'geometry'] = transform(project,pt)
        pointsGDF.loc[j,'distance'] = np.sqrt(x*x+y*y)
    return pointsGDF

def add_north_arrow(ax,
               pos='upper right',
               rel_shift=[None,None],
               shift=[0,0],
               color='k',
               label='N',
               zorder=11,
               fontsize=20):
    validPos ={'upper right',1,
                        'upper left',2,
                        'lower left',3,
                        'lower right',4,
                        'right',5,
                        'center left',6,
                        'center right',7,
                        'lower center',8,
                        'upper center',9,
                        'center',10}
    if pos not in validPos:
        raise ValueError('in scaleBar: pos must be one of %r.' % validPos)
    
    x0, x1, y0, y1 = ax.axis()
    w = (x1-x0)
    h = (y1-y0)
    
    if pos in {'upper right',1}:
        px0 = x1
        py0 = y1
        if rel_shift[0] is None:
            rel_shift=[-0.1,-0.1]
    if pos in {'upper left', 2}:
        px0 = x0
        py0 = y1
        if rel_shift[0] is None:
            rel_shift=[0.1,-0.1]
    if pos in {'lower left', 3}:
        px0 = x0
        py0 = y0
        if rel_shift[0] is None:
            rel_shift=[0.1,0.1]
    if pos in {'lower right',4}:
        px0 = x1
        py0 = y0
        if rel_shift[0] is None:
            rel_shift=[-0.1,0.1]
    if pos in {'right','center right',5,7}:
        px0 = x1
        py0 = y0+0.5*h
    if pos in {'center left',6}:
        px0 = x0
        py0 = y0+0.5*h
    if pos in {'lower center',8}:
        px0 = x0+0.5*w
        py0 = y0
    if pos in {'upper center',9}:
        px0 = x0+0.5*w
        py0 = y1
    if pos in {'center',10}:
        px0 = x0+0.5*w
        py0 = y0+0.5*h
    
    if rel_shift[0] is None: rel_shift[0]=0
    if rel_shift[1] is None: rel_shift[1]=0
    
    px0 += rel_shift[0]*w + shift[0]
    py0 += rel_shift[1]*h + shift[1]
    
    deltaX = 0.03*w
    deltaY = 0.03*h
    
    if len(color)==2:
        colL = color[0]
        colR = color[1]
    else:
        colL = color
        colR = color
    
    arrowPatches = []
    arrowL = np.empty([4,2])
    arrowR = np.empty([4,2])
    arrowL[0,:] = [px0,py0]
    arrowR[0,:] = [px0,py0]
    arrowL[1,:] = [px0,py0+deltaY]
    arrowR[1,:] = [px0,py0+deltaY]
    arrowL[2,:] = [px0-deltaX,py0-0.5*deltaY]
    arrowR[2,:] = [px0+deltaX,py0-0.5*deltaY]
    arrowL[3,:] = arrowL[0,:]
    arrowR[3,:] = arrowR[0,:]
    arrowL_patch = mpatches.Polygon(arrowL,closed=True,facecolor=colL)
    arrowR_patch = mpatches.Polygon(arrowR,closed=True,facecolor=colR)
    arrowPatches.append(arrowL_patch)
    arrowPatches.append(arrowR_patch)
    p = PatchCollection(arrowPatches,match_original=True,zorder=zorder)
    
    ax.add_collection(p)
    Ntext = 'N'
    ntext = ax.text(px0,py0+1.1*deltaY,Ntext,ha='center',fontsize=fontsize,color=color,zorder=zorder)
    return (p,ntext)

def add_scale_bar(ax,
                  pos='lower left',
                  rel_shift=[None,None],
                  shift=[0,0],
                  segments=2,
                  width=None,
                  zorder=11,
                  frame=True,
                  fontsize=6,
                  facecolor=[0.7,0.7,0.7],
                  frame_alpha=1):
    
    validPos ={'upper right',1,
                        'upper left',2,
                        'lower left',3,
                        'lower right',4,
                        'right',5,
                        'center left',6,
                        'center right',7,
                        'lower center',8,
                        'upper center',9,
                        'center',10}
    if pos not in validPos:
        raise ValueError('in scaleBar: pos must be one of %r.' % validPos)
    
    fig = ax.figure

    x0, x1, y0, y1 = ax.axis()
    w = (x1-x0)
    h = (y1-y0)
    
    if width is None:
        width = (ax.get_xticks()[1]-ax.get_xticks()[0])/5
    width = round(width/1000)*1000
    width = max(1000,width)
    height = width/5
    
    if pos in {'upper right',1}:
        px0 = x1
        py0 = y1
        if rel_shift[0] is None:
            rel_shift=[-0.25,-0.05]
    if pos in {'upper left', 2}:
        px0 = x0
        py0 = y1
        if rel_shift[0] is None:
            rel_shift=[0.05,-0.05]
    if pos in {'lower left', 3}:
        px0 = x0
        py0 = y0
        if rel_shift[0] is None:
            rel_shift=[0.05,0.05]
    if pos in {'lower right',4}:
        px0 = x1 - (segments+1)*width
        py0 = y0
        if rel_shift[0] is None:
            rel_shift=[-0.05,0.05]
    if pos in {'right','center right',5,7}:
        px0 = x1
        py0 = y0+0.5*h
    if pos in {'center left',6}:
        px0 = x0
        py0 = y0+0.5*h
    if pos in {'lower center',8}:
        px0 = x0+0.5*w
        py0 = y0
    if pos in {'upper center',9}:
        px0 = x0+0.5*w
        py0 = y1
    if pos in {'center',10}:
        px0 = x0+0.5*w
        py0 = y0+0.5*h
    
    if rel_shift[0] is None: rel_shift[0]=0
    if rel_shift[1] is None: rel_shift[1]=0
    
    px0 += rel_shift[0]*w + shift[0]
    py0 += rel_shift[1]*h + shift[1]
    
    segs = [[x*width,0] for x in range(0,segments+1)]
    pts = _geodesic_point(_ax_gdf(ax).centroid,segs)
    
    xoff = pts.loc[0,'geometry'].x - px0
    yoff = pts.loc[0,'geometry'].y - py0
    
    rect = []
    col = ['k','w']
    texts=[]
    for j in range(1,segments+1):
        anchor_x = pts.loc[j-1,'geometry'].x - xoff
        anchor_y = pts.loc[0,'geometry'].y - yoff
        this_w = pts.loc[j,'geometry'].x - xoff - anchor_x
        rect.append(mpatches.Rectangle((anchor_x,anchor_y),this_w,height,fc=col[0],ec=None,joinstyle='miter'))
        col.reverse()
        scale_text = r'{:d}'.format(int(segs[j][0]/1000))
        if j==segments:
            texts.append(ax.text(anchor_x+this_w,py0-0.02*h,scale_text,fontsize=fontsize,horizontalalignment='center',zorder=zorder))
        else:
            if j==1:
                texts.append(ax.text(anchor_x,py0-0.02*h,r'{:d}'.format(0),fontsize=fontsize,horizontalalignment='center',zorder=zorder))
            texts.append(ax.text(anchor_x+this_w,py0-0.02*h,scale_text,fontsize=fontsize,horizontalalignment='center',zorder=zorder))
    plt.draw()
    inv = ax.transData.inverted()
    bb = inv.transform(texts[-1].get_tightbbox(fig.canvas.get_renderer()).extents)
    texts.append(ax.text(bb[2],py0-0.02*h,r'  km',fontsize=fontsize,horizontalalignment='left',zorder=zorder))
    plt.draw()
    bb1 = inv.transform(texts[-1].get_tightbbox(fig.canvas.get_renderer()).extents)
    
    if frame:
        bbs = np.array([list(r.get_bbox().extents) for r in rect])
        scale_x0 = min(bbs[:,0])
        scale_x1 = max(max(bbs[:,2]),bb1[2])
        scale_y0 = min(bbs[:,1])
        scale_y1 = max(bbs[:,3])
        
        barframe = mpatches.Rectangle((scale_x0-0.02*w,scale_y0-0.02*h),(scale_x1-scale_x0+0.03*w),(scale_y1-scale_y0+0.04*h),fc=facecolor,alpha=frame_alpha,ec=None,zorder=zorder-1)
        ax.add_patch(barframe)
    else:
        barframe = None
    
    p = PatchCollection(rect,zorder=zorder,match_original=True)
    
    ax.add_collection(p)
    
    return (p, barframe)

def _ax_poly(ax,pad_x=0,pad_y=0):
    x0,x1,y0,y1 = ax.axis()
    return Polygon([(x0-pad_x,y0-pad_y),(x0-pad_x,y1+pad_y),(x1+pad_x,y1+pad_y),(x1+pad_x,y0-pad_y)])

def _ax_gdf(ax,crs=webmerc['init'], pad_x=0, pad_y=0):
    axDF = gpd.GeoDataFrame()
    axDF.loc[0,'geometry'] = _ax_poly(ax,pad_x=pad_x,pad_y=pad_y)
    axDF.crs = crs
    return axDF

def _ax_bbox(ax,crs=webmerc['init']):
    axGDF = _ax_gdf(ax,crs=crs).to_crs(wgs84['init'])
    north = axGDF.bounds.maxy.values[0]
    south = axGDF.bounds.miny.values[0]
    east = axGDF.bounds.maxx.values[0]
    west = axGDF.bounds.minx.values[0]
    return north, south, east, west

def _get_osm_roads(ax, crs=webmerc['init']):
    north, south, east, west = _ax_bbox(ax, crs=crs)
    G = osm.graph_from_bbox(north, south, east, west,
                            network_type='drive_service')
    nodes, edges = osm.graph_to_gdfs(G)
    roads = edges.to_crs(crs)
    return roads

def _get_osm_buildings(ax, crs=webmerc['init']):
    north, south, east, west = _ax_bbox(ax, crs=crs)
    fp=osm.geometries.geometries_from_bbox(north=north,
                                           south=south,
                                           east=east,
                                           west=west,
                                           tags={'building':True})
    buildings = fp.to_crs(crs)
    return buildings

def add_osm(ax,
            roads=True,
            buildings=False,
            road_color='black',
            building_color='gray'):
    if roads:
        roads = _get_osm_roads(ax)
        roads.plot(ax=ax, color=road_color)
    if buildings:
        buildings = _get_osm_buildings(ax)
        buildings.plot(ax=ax, color=building_color)
    return


def add_POI_marker(POI,
                   ax=None,
                   marker='.',
                   markersize=8,
                   color='k',
                   edgecolor='k',
                   linewidth=0,
                   label=None,
                   zorder=0,
                   avoid_overlap=None):
    
    x0, x1, y0, y1 = ax.axis()
    axPoly = _ax_poly(ax)
    
    if label is not None:
        if label not in POI.columns:
            raise RuntimeError('in add_POI_marker plot, label must be in the POI columns')
    
    POIax = POI[POI.geometry.intersects(axPoly)]
    
    for i, p in POI.iterrows():
        px = p.geometry.x
        py = p.geometry.y
        if x0<=px<=x1 and y0<=py<=y1:
            ax.scatter(px, py, markersize, marker=marker, facecolor=color, edgecolors=edgecolor, linewidths=linewidth, zorder=zorder)
            ax.annotate(p[label], xy=(px,py), xytext=(3, 3), textcoords="offset points")
    
    return


from numpy import ceil, sqrt


def plot_rowscols(panels):
    rows = int(ceil(sqrt(panels)))
    cols = int(ceil(panels/rows))
    return rows, cols


def pad_window(window, height, width, pad=1):
    row_start = max(window.row_off - pad, 0)
    row_stop = min(window.row_off + window.height - 1 + pad, height - 1)
    rows = row_stop - row_start + 1
    col_start = max(window.col_off - pad, 0)
    col_stop = min(window.col_off + window.width - 1 + pad, width - 1)
    columns = col_stop - col_start + 1
    return {'row_start': row_start, 'row_stop': row_stop, 'rows': rows,
            'col_start': col_start, 'col_stop': col_stop, 'columns': columns}



import datetime
import re


def is_valid_datetime(input_datetime):
    datetime_fmt = re.compile(r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}')
    return datetime_fmt.match(input_datetime) is not None


def string_to_datetime(datetime_string):
    if not is_valid_datetime(datetime_string):
        raise ValueError('{} is not a valid datetime string.  Required format is yyyy-mm-dd hh:mm'.format(datetime_string))
    (date, time) = datetime_string.split(' ', 2)
    [year, month, day] = [int(x) for x in date.split('-', 3)]
    [hour, minute] = [int(x) for x in time.split(':', 2)]

    output_datetime = (datetime.datetime(year=year, month=month, day=day,
                                             hour=hour, minute=0)
                            + datetime.timedelta(hours=minute//30))

    return output_datetime